import struct
import ipaddress
import iCOMOX_messages
import helpers

cMEMMAP_VERSION = 1
cMEMMAP_USED_PAGES = 20
cMEMMAP_SIZE = iCOMOX_messages.M24C64_PAGE_BYTE_SIZE * cMEMMAP_USED_PAGES

list_iCOMOX_PartNumber = ["SRT-COMOX-SM", "SRT-iCOMOX-NBIOT-SEC", "SRT-ICOMOX-POE-SEC"]

CipherSuitsArr = [
    0x0035,  # "TLS_RSA_WITH_AES_256_CBC_SHA",
    0x002F,  # "TLS_RSA_WITH_AES_128_CBC_SHA",
    0x0005,  # "TLS_RSA_WITH_RC4_128_SHA",
    0x0004,  # "TLS_RSA_WITH_RC4_128_MD5",
    0x000A,  # "TLS_RSA_WITH_3DES_EDE_CBC_SHA",
    0x003D,  # "TLS_RSA_WITH_AES_256_CBC_SHA256",
    0xC011,  # "TLS_ECDHE_RSA_WITH_RC4_128_SHA",
    0xC012,  # "TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA",
    0xC013,  # "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA",
    0xC014,  # "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA",
    0xC027,  # "TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256",
    0xC028,  # "TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA384",
    0xC02F,  # "TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256",
    0xFFFF  # "Support all"
]

def ValidPositiveInteger(string):
    return (len(string) > 0) and all(c in "0123456789" for c in string)

def ValidPLMNString(PLMN_string):
    return (len(PLMN_string) >= 5) and (len(PLMN_string) <= 6) and ValidPositiveInteger(PLMN_string)

class cMemMap():
    def __init__(self):
        self.memmap = None
        self.Clear()

    def Clear(self):
        self.memmap = bytearray(b"\xFF") * cMEMMAP_SIZE

    def AssignStringToMemMap(self, position, data, maxlen, fillchar=None):
        if len(data) > maxlen:
            data = data[:maxlen]
        if type(data) is str:
            data = bytearray(data, "utf-8", errors="ignore")
        self.memmap[position:position+len(data)] = data
        if fillchar is not None:
            self.memmap[position+len(data):position + maxlen] = fillchar * (maxlen - len(data))

    def StringFromMemMap(self, position, maxlen, fillchar):
        data = self.memmap[position:position+maxlen]
        index = data.find(fillchar)
        if index >= 0:
            data = data[:index]
        return str(data, "utf-8", errors="ignore")

    def WinToMemMap_Common(self, MemMapVersion, BoardType, BoardVersionMajor, BoardVersionMinor, PartNumber, SerialNumber, Name, ProductionSupport=False):
        if ProductionSupport:
            if len(SerialNumber) != 11:
                raise Exception("Serial number should be 11 characters")
            self.memmap[0] = MemMapVersion
            self.memmap[1] = int(BoardType)
            self.memmap[2] = int(BoardVersionMajor)
            self.memmap[3] = int(BoardVersionMinor)
            self.memmap[4:32] = b"\xFF" * 28

            self.AssignStringToMemMap(position=32, data=PartNumber, maxlen=32, fillchar=b"\xFF")
            self.AssignStringToMemMap(position=64, data=SerialNumber, maxlen=32, fillchar=b"\xFF")
        else:
            self.memmap[0:96] = b"\xFF" * 96
        self.AssignStringToMemMap(position=96, data=Name, maxlen=32, fillchar=b"\xFF")

    def MemMapToWin_Common(self):
        MemMapVersion = int(self.memmap[0])
        BoardType = int(self.memmap[1])
        BoardVersionMajor = int(self.memmap[2])
        BoardVersionMinor = int(self.memmap[3])
        PartNumber = self.StringFromMemMap(position=32, maxlen=32, fillchar=b"\xFF")
        SerialNumber = self.StringFromMemMap(position=64, maxlen=32, fillchar=b"\xFF")
        Name = self.StringFromMemMap(position=96, maxlen=32, fillchar=b"\xFF")
        return MemMapVersion, BoardType, BoardVersionMajor, BoardVersionMinor, PartNumber, SerialNumber, Name

    def WinToMemMap_SSL(self, addr_in_memmap, SSL_Enable, SSL_version, SSL_SecLevel, SSL_CipherSuite, SSL_NegotiateTimeout, SSL_IgnoreLocalTime):
        self.memmap[addr_in_memmap:addr_in_memmap+32] = b"\xFF" * 32
        # self.AssignStringToMemMap(position=addr_in_memmap,    data=SSL_CACertFile,      maxlen=32, fillchar=b"\xFF")
        # self.AssignStringToMemMap(position=addr_in_memmap+32, data=SSL_ClientCertFile,  maxlen=32, fillchar=b"\xFF")
        # self.AssignStringToMemMap(position=addr_in_memmap+64, data=SSL_ClientKeyFile,   maxlen=32, fillchar=b"\xFF")
        self.memmap[addr_in_memmap] = CipherSuitsArr[SSL_CipherSuite] & 0xFF
        self.memmap[addr_in_memmap+1] = CipherSuitsArr[SSL_CipherSuite] >> 8
        self.memmap[addr_in_memmap+2] = SSL_NegotiateTimeout & 0xFF
        self.memmap[addr_in_memmap+3] = SSL_NegotiateTimeout >> 8
        self.memmap[addr_in_memmap+4] = (SSL_version & 0x07) + ((SSL_SecLevel & 0x03) << 4) + ((SSL_IgnoreLocalTime & 0x01) << 6) + ((SSL_Enable & 0x01) << 7)
        self.memmap[addr_in_memmap+5:addr_in_memmap+31] = b"\xFF" * 127

    def MemMapToWin_SSL(self, addr_in_memmap):
        # SSL_CACertFile      = str(self.StringFromMemMap(position=addr_in_memmap,    maxlen=32, fillchar=b"\xFF"), "utf8")
        # SSL_ClientCertFile  = str(self.StringFromMemMap(position=addr_in_memmap+32, maxlen=32, fillchar=b"\xFF"), "utf8")
        # SSL_ClientKeyFile   = str(self.StringFromMemMap(position=addr_in_memmap+64, maxlen=32, fillchar=b"\xFF"), "utf8")
        try:
            SSL_CipherSuite = CipherSuitsArr.index((self.memmap[addr_in_memmap+1] << 8) + self.memmap[addr_in_memmap])
        except:
            SSL_CipherSuite = len(CipherSuitsArr)-1
        finally:
            pass
        SSL_NegotiateTimeout = (self.memmap[addr_in_memmap+3] << 8) + self.memmap[addr_in_memmap+2]
        if SSL_NegotiateTimeout < 10:
            SSL_NegotiateTimeout = 10
        elif SSL_NegotiateTimeout > 300:
            SSL_NegotiateTimeout = 300
        SSL_version = self.memmap[addr_in_memmap+4] & 0x07
        if SSL_version > 4:
            SSL_version = 0
        SSL_SecLevel = (self.memmap[addr_in_memmap+4] >> 4) & 0x03
        if SSL_SecLevel > 2:
            SSL_SecLevel = 0
        SSL_IgnoreLocalTime = 0 != (self.memmap[addr_in_memmap+4] & 0x40)
        SSL_Enable = 0 != (self.memmap[addr_in_memmap+4] & 0x80)
        return SSL_Enable, SSL_version, SSL_SecLevel, SSL_CipherSuite, SSL_NegotiateTimeout, SSL_IgnoreLocalTime

    def WinToMemMap_MQTT(self, addr_in_memmap, MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud):
        self.AssignStringToMemMap(position=addr_in_memmap,      data=MQTT_ServerURL,    maxlen=128, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=addr_in_memmap+32*4, data=MQTT_UserName,     maxlen=32,  fillchar=b"\xFF")
        self.AssignStringToMemMap(position=addr_in_memmap+32*5, data=MQTT_Password,     maxlen=32,  fillchar=b"\xFF")
        self.AssignStringToMemMap(position=addr_in_memmap+32*6, data=MQTT_ClientID,     maxlen=32,  fillchar=b"\xFF")
        self.memmap[addr_in_memmap+32*7] = MQTT_ServerPort & 0xFF
        self.memmap[addr_in_memmap+32*7+1] = (MQTT_ServerPort >> 8) & 0xFF
        self.memmap[addr_in_memmap+32*7+2] = MQTT_Cloud
        self.memmap[addr_in_memmap+32*7+3] = ((MQTT_ProvideUserName & 0x01) << 7) +((MQTT_ProvidePassword & 0x01) << 6)    # flags
        self.memmap[addr_in_memmap+32*7+4] = MQTT_Keepalive & 0xFF
        self.memmap[addr_in_memmap+32*7+5] = (MQTT_Keepalive >> 8) & 0xFF
        self.memmap[addr_in_memmap+32*7+6:addr_in_memmap+32*7+32] = b"\xFF" * 28

    def MemMapToWin_MQTT(self, addr_in_memmap):
        MQTT_ServerURL      = self.StringFromMemMap(position=addr_in_memmap,        maxlen=128, fillchar=b"\xFF")
        MQTT_UserName       = self.StringFromMemMap(position=addr_in_memmap+32*4,   maxlen=32,  fillchar=b"\xFF")
        MQTT_Password       = self.StringFromMemMap(position=addr_in_memmap+32*5,   maxlen=32,  fillchar=b"\xFF")
        MQTT_ClientID       = self.StringFromMemMap(position=addr_in_memmap+32*6,   maxlen=32,  fillchar=b"\xFF")
        MQTT_ServerPort     = (self.memmap[addr_in_memmap+32*7+1] << 8) + self.memmap[addr_in_memmap+32*7]
        MQTT_Cloud          = self.memmap[addr_in_memmap+32*7+2] & 0x01
        MQTT_ProvideUserName = 0 != (self.memmap[addr_in_memmap+32*7+3] & 0x80)
        MQTT_ProvidePassword = 0 != (self.memmap[addr_in_memmap + 32 * 7 + 3] & 0x40)
        MQTT_Keepalive      = (self.memmap[addr_in_memmap+32*7+5] << 8) + self.memmap[addr_in_memmap+32*7+4]
        if MQTT_Cloud > 0:
            MQTT_Cloud = 0
        return MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud

    def WinToMemMap_TCP(self, addr_in_memmap, TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec):
        # Listen IP address
        # if TCP_IPv6:
        #     ip = ipaddress.IPv6Address(TCP_ServerURL).packed
        # else:
        #     ip = ipaddress.IPv4Address(TCP_ServerURL).packed
        # self.AssignStringToMemMap(position=addr_in_memmap, data=ip, maxlen=16, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=addr_in_memmap, data=TCP_ServerURL, maxlen=128, fillchar=b"\xFF")
        self.memmap[addr_in_memmap+128] = TCP_ServerPort & 0xFF
        self.memmap[addr_in_memmap+129] = ((TCP_ServerPort >> 8) & 0xFF)
        self.memmap[addr_in_memmap+130] = TCP_delayBeforeReconnectingSec & 0xFF
        # self.memmap[addr_in_memmap+131] = TCP_connectAttemptsCount & 0xFF
        # self.memmap[addr_in_memmap+132] = 0 #TCP_IPv6 & 0x01

    def MemMapToWin_TCP(self, addr_in_memmap):
        TCP_ServerURL = self.StringFromMemMap(position=addr_in_memmap, maxlen=128, fillchar=b"\xFF")
        TCP_ServerPort = (self.memmap[addr_in_memmap+129] << 8) + self.memmap[addr_in_memmap+128]
        TCP_delayBeforeReconnectingSec = self.memmap[addr_in_memmap+130]
        # TCP_connectAttemptsCount = self.memmap[addr_in_memmap+131]
        # TCP_Flags = self.memmap[addr_in_memmap+132]
        return TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec

    def WinToMemMap_Backbone(self, addr_in_memmap, BackboneProtocol, TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud):
        self.memmap[addr_in_memmap:addr_in_memmap+9*32] = b"\xFF" * (9*32)
        if BackboneProtocol == iCOMOX_messages.cBACKBONE_PROTOCOL_TCP:
            self.WinToMemMap_TCP(addr_in_memmap=addr_in_memmap, TCP_ServerURL=TCP_ServerURL, TCP_ServerPort=TCP_ServerPort, TCP_delayBeforeReconnectingSec=TCP_delayBeforeReconnectingSec)
        elif BackboneProtocol == iCOMOX_messages.cBACKBONE_PROTOCOL_MQTT:
            self.WinToMemMap_MQTT(addr_in_memmap=addr_in_memmap, MQTT_ServerURL=MQTT_ServerURL, MQTT_UserName=MQTT_UserName, MQTT_Password=MQTT_Password, MQTT_ClientID=MQTT_ClientID, MQTT_ServerPort=MQTT_ServerPort, MQTT_ProvideUserName=MQTT_ProvideUserName, MQTT_ProvidePassword=MQTT_ProvidePassword, MQTT_Keepalive=MQTT_Keepalive, MQTT_Cloud=MQTT_Cloud)

    def MemMapToWin_Backbone(self, addr_in_memmap, BackboneProtocol):
        TCP_ServerURL = TCP_ServerPort = TCP_delayBeforeReconnectingSec = MQTT_ServerURL = MQTT_UserName = MQTT_Password = MQTT_ClientID = MQTT_ServerPort = MQTT_ProvideUserName = MQTT_ProvidePassword = MQTT_Keepalive = MQTT_Cloud = None
        if BackboneProtocol == iCOMOX_messages.cBACKBONE_PROTOCOL_TCP:
            TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec = self.MemMapToWin_TCP(addr_in_memmap=addr_in_memmap)
        elif BackboneProtocol == iCOMOX_messages.cBACKBONE_PROTOCOL_MQTT:
            MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud = self.MemMapToWin_MQTT(addr_in_memmap=addr_in_memmap)
        return TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud

    def WinToMemMap_NBIOT(self, \
          UseDefSimConfiguration, AccessTechnologiesBitmask, GsmBandsBitmask, LteCatM1BandsBitmask, LteCatNB1BandsBitmask, ScanOrder, ServiceDomain, \
          ManualOperatorSelection, PLMN, \
          EnableRoaming, ContextType, Authentication, ApnAccessName, ApnUser, ApnPassword, \
          SSL_Enable, SSL_version, SSL_SecLevel, SSL_CipherSuite, SSL_NegotiateTimeout, SSL_IgnoreLocalTime, \
          BackboneProtocol, \
          TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
          MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud
          ):
        self.memmap[32*4:32*(cMEMMAP_USED_PAGES)] = b"\xFF" * (32*(cMEMMAP_USED_PAGES-4))
        # flags
        self.memmap[32*4] = (UseDefSimConfiguration & 0x01) + ((ManualOperatorSelection & 0x01) << 1) + ((EnableRoaming & 0x01) << 2) + ((ServiceDomain & 0x01) << 3)   # flags.flags
        self.memmap[32*4+1] = (AccessTechnologiesBitmask & 0x07) + ((ScanOrder & 0x07) << 3)                                                                            # flags.Access
        self.memmap[32*4+2] = (Authentication & 0x03) + ((ContextType & 0x03) << 2)                                                                                     # flags.Context
        self.memmap[32*4+3] = BackboneProtocol & 0x01                                                                                                                   # flags.protocols

        # PLMN
        if not ValidPLMNString(PLMN):
            PLMN = ""
        self.AssignStringToMemMap(position=32*4+4, data=PLMN, maxlen=6, fillchar=b"\xFF")

        # GSM bands bitmask
        self.memmap[32*4+10] = GsmBandsBitmask & 0xFF
        self.memmap[32*4+11] = (GsmBandsBitmask >> 8) & 0xFF

        # LTE cat M1 bands bitmask
        self.memmap[32*4+12] = LteCatM1BandsBitmask & 0xFF
        self.memmap[32*4+13] = (LteCatM1BandsBitmask >> 8) & 0xFF

        # LTE cat NB1 bands bitmask
        self.memmap[32*4+14] = LteCatNB1BandsBitmask & 0xFF
        self.memmap[32*4+15] = (LteCatNB1BandsBitmask >> 8) & 0xFF

        # Dummy
        # self.memmap[32*4+16:32*4+32] = b"\xFF" * 18

        self.AssignStringToMemMap(position=32*5, data=ApnAccessName, maxlen=32, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=32*6, data=ApnUser, maxlen=32, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=32*7, data=ApnPassword, maxlen=32, fillchar=b"\xFF")

        self.WinToMemMap_SSL(addr_in_memmap=32*8, SSL_Enable=SSL_Enable, SSL_version=SSL_version, SSL_SecLevel=SSL_SecLevel, SSL_CipherSuite=SSL_CipherSuite, SSL_NegotiateTimeout=SSL_NegotiateTimeout, SSL_IgnoreLocalTime=SSL_IgnoreLocalTime)

        self.WinToMemMap_Backbone(addr_in_memmap=32*9, BackboneProtocol=BackboneProtocol, \
              TCP_ServerURL=TCP_ServerURL, TCP_ServerPort=TCP_ServerPort, TCP_delayBeforeReconnectingSec=TCP_delayBeforeReconnectingSec, \
              MQTT_ServerURL=MQTT_ServerURL, MQTT_UserName=MQTT_UserName, MQTT_Password=MQTT_Password, MQTT_ClientID=MQTT_ClientID, MQTT_ServerPort=MQTT_ServerPort, MQTT_ProvideUserName=MQTT_ProvideUserName, MQTT_ProvidePassword=MQTT_ProvidePassword, MQTT_Keepalive=MQTT_Keepalive, MQTT_Cloud=MQTT_Cloud)

    def MemMapToWin_NBIOT(self):
        # flags.flags
        UseDefSimConfiguration = 0 != (self.memmap[32*4] & 0x01)
        ManualOperatorSelection = 0 != (self.memmap[32*4] & 0x02)
        EnableRoaming = 0 != (self.memmap[32*4] & 0x04)
        ServiceDomain = (self.memmap[32*4] >> 3) & 0x01
        # flags.Access
        AccessTechnologiesBitmask = self.memmap[32*4+1] & 0x07
        ScanOrder = (self.memmap[32*4+1] >> 3) & 0x07
        if ScanOrder > 5:
            ScanOrder = 0

        # flags.Context
        Authentication = self.memmap[32*4+2] & 0x03
        ContextType = (self.memmap[32*4+2] >> 2) & 0x03
        if ContextType > 2:
            ContextType = 0

        # flags.protocols
        BackboneProtocol = self.memmap[32*4+3] & 0x01
        if BackboneProtocol > 0:
            BackboneProtocol = 0

        # PLMN
        PLMN = self.StringFromMemMap(position=32*4+4, maxlen=6, fillchar=b"\xFF")
        if not ValidPLMNString(PLMN_string=PLMN):
            PLMN = ""
        # MNC = ((self.memmap[32*4+5] & 0x7F) << 8) + self.memmap[32*4+4]
        # MNCHas3Digits = 0 != (self.memmap[32*4+5] & 0x80)
        # MCC = (self.memmap[32*4+7] << 8) + self.memmap[32*4+6]

        # Bands bitmask
        GsmBandsBitmask = (self.memmap[32*4+11] << 8) + self.memmap[32*4+10]
        LteCatM1BandsBitmask = (self.memmap[32*4+13] << 8) + self.memmap[32*4+12]
        LteCatNB1BandsBitmask = (self.memmap[32*4+15] << 8) + self.memmap[32*4+14]

        ApnAccessName   = self.StringFromMemMap(position=32*5, maxlen=32, fillchar=b"\xFF")
        ApnUser         = self.StringFromMemMap(position=32*6, maxlen=32, fillchar=b"\xFF")
        ApnPassword     = self.StringFromMemMap(position=32*7, maxlen=32, fillchar=b"\xFF")

        SSL_Enable, SSL_version, SSL_SecLevel, SSL_CipherSuite, SSL_NegotiateTimeout, SSL_IgnoreLocalTime = self.MemMapToWin_SSL(addr_in_memmap=32*8)

        TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
        MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud = self.MemMapToWin_Backbone(addr_in_memmap=32*9, BackboneProtocol=BackboneProtocol)

        return  UseDefSimConfiguration, AccessTechnologiesBitmask, GsmBandsBitmask, LteCatM1BandsBitmask, LteCatNB1BandsBitmask, ScanOrder, ServiceDomain, \
                ManualOperatorSelection, PLMN,  \
                EnableRoaming, ContextType, Authentication, ApnAccessName, ApnUser, ApnPassword, \
                SSL_Enable, SSL_version, SSL_SecLevel, SSL_CipherSuite, SSL_NegotiateTimeout, SSL_IgnoreLocalTime, \
                BackboneProtocol, \
                TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
                MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud

    # def MemMapToWin_NBIOT(self):
    #     IPv6 = 0 != (self.memmap[148] & 1)
    #     TLS = 0 != (self.memmap[148] & 2)
    #     if IPv6:
    #         TcpServerIpAddr = ipaddress.IPv6Address(bytes(self.memmap[128:128+16])).exploded
    #     else:
    #         TcpServerIpAddr = ipaddress.IPv4Address(bytes(self.memmap[128:128+4])).exploded
    #     [TcpServerPort] = struct.unpack_from("<H", self.memmap, 144)
    #     connectAttemptsCount = self.memmap[146]
    #     delayBeforeConnectingSec = self.memmap[147]
    #     ApnAccessName = self.StringFromMemMap(position=32*5, maxlen=32, fillchar=b"\xFF")
    #     ApnUser = self.StringFromMemMap(position=32*6, maxlen=32, fillchar=b"\xFF")
    #     ApnPassword = self.StringFromMemMap(position=32*7, maxlen=32, fillchar=b"\xFF")
    #     return TcpServerIpAddr, TcpServerPort, connectAttemptsCount, delayBeforeConnectingSec, IPv6, TLS, ApnAccessName, ApnUser, ApnPassword

    def WinToMemMap_POE(self, \
                        StaticIpAddr, StaticMaskAddr, StaticDnsServerAddr, StaticGatewayAddr, IPv6, IPv6LocalLinkOnly, DHCP,  \
                        BackboneProtocol, \
                        TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
                        MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud):
        # TCP/IP interface
        clientIP = helpers.stringToIpAddress(StaticIpAddr)
        clientIP = bytearray() if clientIP is None else clientIP.packed
        Mask = helpers.stringToIpAddress(StaticMaskAddr)
        Mask = bytearray() if Mask is None else Mask.packed
        DnsServerAddr = helpers.stringToIpAddress(StaticDnsServerAddr)
        DnsServerAddr = bytearray() if DnsServerAddr is None else DnsServerAddr.packed
        GatewayIP = helpers.stringToIpAddress(StaticGatewayAddr)
        GatewayIP = bytearray() if GatewayIP is None else GatewayIP.packed
        self.AssignStringToMemMap(position=128, data=clientIP, maxlen=16, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=144, data=Mask, maxlen=16, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=160, data=DnsServerAddr, maxlen=16, fillchar=b"\xFF")
        self.AssignStringToMemMap(position=176, data=GatewayIP, maxlen=4, fillchar=b"\xFF")

        # Flags
        self.memmap[192] = 0
        if IPv6:
            self.memmap[192] |= 1
        if IPv6LocalLinkOnly:
            self.memmap[192] |= 2
        # if TLS:
        #     self.memmap[192] |= 4
        if DHCP:
            self.memmap[192] |= 8
        self.memmap[193] = BackboneProtocol & 0x01

        self.WinToMemMap_Backbone(addr_in_memmap=256, BackboneProtocol=BackboneProtocol, \
              TCP_ServerURL=TCP_ServerURL, TCP_ServerPort=TCP_ServerPort, TCP_delayBeforeReconnectingSec=TCP_delayBeforeReconnectingSec, \
              MQTT_ServerURL=MQTT_ServerURL, MQTT_UserName=MQTT_UserName, MQTT_Password=MQTT_Password, MQTT_ClientID=MQTT_ClientID, MQTT_ServerPort=MQTT_ServerPort, MQTT_ProvideUserName=MQTT_ProvideUserName, MQTT_ProvidePassword=MQTT_ProvidePassword, MQTT_Keepalive=MQTT_Keepalive, MQTT_Cloud=MQTT_Cloud)

    def MemMapToWin_POE(self):
        IPv6 = 0 != (self.memmap[192] & 1)
        IPv6LocalLinkOnly = 0 != (self.memmap[192] & 2)
        # TLS = 0 != (self.memmap[192] & 4)
        DHCP = 0 != (self.memmap[192] & 8)

        if IPv6:
            StaticIpAddr = ipaddress.IPv6Address(bytes(self.memmap[128:144])).exploded
            StaticMaskAddr = ipaddress.IPv6Address(bytes(self.memmap[144:160])).exploded
            StaticDnsServerAddr = ipaddress.IPv6Address(bytes(self.memmap[160:176])).exploded
            StaticGatewayAddr = ipaddress.IPv6Address(bytes(self.memmap[176:192])).exploded
        else:
            StaticIpAddr = ipaddress.IPv4Address(bytes(self.memmap[128:132])).exploded
            StaticMaskAddr = ipaddress.IPv4Address(bytes(self.memmap[144:148])).exploded
            StaticDnsServerAddr = ipaddress.IPv4Address(bytes(self.memmap[160:164])).exploded
            StaticGatewayAddr = ipaddress.IPv4Address(bytes(self.memmap[176:180])).exploded

        BackboneProtocol = self.memmap[193] & 0x01
        BackboneProtocol = iCOMOX_messages.cBACKBONE_PROTOCOL_TCP   # Remove in 2.8.1

        TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
        MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud = self.MemMapToWin_Backbone(addr_in_memmap=256, BackboneProtocol=BackboneProtocol)

        return  StaticIpAddr, StaticMaskAddr, StaticDnsServerAddr, StaticGatewayAddr, IPv6, IPv6LocalLinkOnly, DHCP, \
                BackboneProtocol, \
                TCP_ServerURL, TCP_ServerPort, TCP_delayBeforeReconnectingSec, \
                MQTT_ServerURL, MQTT_UserName, MQTT_Password, MQTT_ClientID, MQTT_ServerPort, MQTT_ProvideUserName, MQTT_ProvidePassword, MQTT_Keepalive, MQTT_Cloud



"""
This module implements the Gateway Global State.

The Gateway Global State is responsible to hold:
    - The Gateway States Machine.
    - The Persistence of the system.
    - Any kind of information that needs to be persisted when the
        Gateway transitions between states.
"""
# ============================ imports =========================================
from typing import List, Dict
from txp.common.configuration import GatewayConfig
from txp.common.edge import EdgeDescriptor, MachineMetadata
from txp.devices.gateway_states_enums import *
from dataclasses import dataclass
from txp.common.config import settings
import datetime
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


@dataclass
class GatewayGlobalState:
    """The Gateway Global state class. This dataclass defines Gateway attributes that are
    collected in the configuration process, and needs to be persisted.
    """

    gateway_conf: GatewayConfig
    """gateway_config: The current GatewayConfig of the Gateway."""
    machines_configurations: List[MachineMetadata]
    """machines_configurations: The logical configurations for the Machines monitored by the 
    Gateway"""
    edges_configurations: List[EdgeDescriptor]
    """edges_configurations: The logical configurations for the Edges configured to work with 
    this Gateway"""
    state: GatewayStates
    """state: The current state value of the Gateway"""
    status: ...
    """status: An status member defined in the status enumeration for the current state"""
    packages_collected: int
    """packages_collected: an attribute to track how many packages has been collected by the Gateway."""
    collected_packages_by_edge: Dict[str, int]
    """collected_package_by_edge: A table to count the number of packages collected by edges using the serial as key"""
    packages_sent: int
    """packages_sent: an attribute to track how many packages has been sent to the cloud"""
    boot_time: datetime
    """boot_time: The boot time of the Gateway instance."""
    last_configuration_id: str
    """last_configuration_id: The last configuration ID received in the Gateway."""



import ray
from fastapi import FastAPI, status
from ray import serve
from txp.ml.prediction_service.service_manager.prediction_manager import PredictionManager
import time
import logging
import argparse
from fastapi.responses import JSONResponse

app = FastAPI()
ray.init()

"""Prediction service.

Args:
    backup_collection_name: Firestore collection name where all backups are going to be stored.

    dataset: Bigquery dataset name where data is going to be read from, this is where telemetry data is stored.

    events_and_states_dataset: Bigquery dataset where events and states are going to be stored.

    subscription_id: PubSub subscription id where signal messages are going to come from, this is the input
    of the service.

    notifications_topic: PubSub topic id where notifications are going to be sent.

    credentials_path: GCS credentials local file path.

"""


@serve.deployment(route_prefix="/prediction", num_replicas=1)
@serve.ingress(app)
class PredictionService:
    def __init__(self, backup_collection_name, dataset, events_and_states_dataset, subscription_id, notifications_topic,
                 _credentials_str, log_level=logging.INFO):

        logging.basicConfig(level=log_level)
        logging.info(f"Initialization started for deployment {self.__class__.__name__}")

        self.__credentials_str = _credentials_str
        self.__log_level = log_level
        self.__prediction_mgr: PredictionManager = \
            PredictionManager.remote(dataset, subscription_id, backup_collection_name, events_and_states_dataset,
                                     notifications_topic, self.__credentials_str, self.__log_level)

        logging.info(f"Initialization finished successfully for deployment class {self.__class__.__name__}")

    @app.get("/")
    def root(self):
        return JSONResponse(
            status_code=status.HTTP_200_OK,
            content={"message": "Training service is up and running"},
        )


if __name__ == "__main__":
    serve.start()

    parser = argparse.ArgumentParser()
    parser.add_argument("--backup_collection_name", help="Backup collection name",
                        default='ray_service_backup_test')
    parser.add_argument("--dataset", help="Dataset name where data is stored",
                        default="telemetry_test")
    parser.add_argument("--events_and_states_dataset",
                        help="Dataset name where events and state are going to be stored",
                        default="ml_events_and_states_test")
    parser.add_argument("--subscription_id", help="pubsub subscription id where messages are going to come from",
                        default="txp-model-serving-signals-test-sub")
    parser.add_argument("--notifications_topic", help="pubsub topic id where notifications are going to be sent",
                        default="txp-notifications-local-test")
    parser.add_argument("--credentials_path", help="Google cloud credentials path",
                        default="./credentials/pub_sub_to_bigquery_credentials.json")
    known_args, _ = parser.parse_known_args()

    with open(known_args.credentials_path, 'r') as file:
        credentials_str = file.read().replace('\n', '')

    PredictionService.deploy(known_args.backup_collection_name, known_args.dataset,
                             known_args.events_and_states_dataset, known_args.subscription_id,
                             known_args.notifications_topic, credentials_str, logging.INFO)

    while True:
        time.sleep(600)



from typing import List

from txp.ml.prediction_service.state_managers import state_manager as sm
import ray
import logging
from txp.common.ml.tasks import AssetState, AssetStateCondition


class BeerShowroomAsset(sm.StateManager):
    states_map = {
        "Present": "OPTIMAL",
        "Good": "OPERATIONAL CONDITION",
        "Bad": "OPERATIONAL CONDITION",
        "NotPresent": "CRITICAL"
    }

    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                 credentials_str, log_level=logging.INFO):
        logging.basicConfig(level=log_level)
        super().__init__(asset_id, events_and_states_dataset, backup_collection_name, notifications_topic,
                         credentials_str)
        self.tasks = tasks
        self.states_count = 3
        self.non_persisted_count = 0

    def get_state(self, events):
        """
        Args:

            events: List of events produced by the tasks associated to this
                Asset.

        Return:
            A state object (typing.Dict)
        """
        recv_events: List[str] = list(map(
            lambda event_obj: event_obj['event'],
            events
        ))

        beer_present: bool = any(
            map(
                lambda event_str: event_str == 'PresentBeer',
                recv_events
            )
        )

        ids: List[str] = list(map(
            lambda event_obj: event_obj['event_id'],
            events
        ))

        logging.info(f"{self.__class__.__name__} received new events for asset {events[0]['asset_id']}. "
                     f"Received events: {recv_events}")

        state = AssetState(events=ids, asset_id=self.asset_id,
                           condition=AssetStateCondition.OPTIMAL if beer_present else AssetStateCondition.CRITICAL,
                           observation_timestamp=events[0]["observation_timestamp"],
                           tenant_id=events[0]["tenant_id"],
                           partition_timestamp=events[0]["partition_timestamp"])

        logging.info(f"{self.__class__.__name__} computed new state for asset {state.asset_id}. "
                     f"New state value: {state.condition.value} ")
        return state

    def process_new_state(self, state, gateway_task_id):
        logging.info(f"{self.__class__.__name__} Requested to process new state")
        last_states = self.get_last_states(self.states_count, gateway_task_id)
        # number of states needed to decide to write in the table
        if not len(last_states):
            # TODO: Change to log usage when available
            logging.info(f"{self.__class__.__name__} new state for a new Gateway Task ID. "
                         f"Persist into Database...")
            self.insert_state_to_bigquery(state)
            return

        # if new state is different that the immediate previous one, then write
        if state["condition"] != last_states[0]["condition"]:
            logging.info(f"{self.__class__.__name__} new state different than the previous one. "
                         f"Persist into Database...")
            self.insert_state_to_bigquery(state)
            return

        if self.non_persisted_count == self.states_count:
            logging.info(f"{self.__class__.__name__} Maximum number of non persisted states reached. "
                         f"Persist into Database...")
            self.insert_state_to_bigquery(state)
            self.non_persisted_count = 0
            return

        self.non_persisted_count += 1


@ray.remote
class RayBasicShowroomBeerAsset(BeerShowroomAsset):
    def __init__(self, asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                 credentials_str, log_level=logging.INFO):
        super().__init__(asset_id, tasks, events_and_states_dataset, backup_collection_name, notifications_topic,
                         credentials_str, log_level)



import configparser
import common.helpers

__VERSION_CODE__ = 0x020800
__VERSION__ = "{}.{}.{}".format(__VERSION_CODE__ >> 16, (__VERSION_CODE__ >> 8) & 0xFF, __VERSION_CODE__ & 0xFF)

__FAULT_DETECTION_ALGORITHM_INDEX__ = 1

__REST_SUPPORT__ = False

__PRODUCTION_SUPPORT__ = True
__ADXL1002_GRAPH_SUPPORT__ = False

# Configuration variables that should be read from an external configuration file
BMM150_quantile_for_noise_floor_estimator   = 0.25
BMM150_minimum_SNR_for_speed_detection_dB   = 20   # 20dB of minimum threshold above the noise floor to detect synchronous speed

ASYNC_MOTOR_number_of_poles_pairs           = 2
ASYNC_MOTOR_slip_factor_percentages         = 0.0

IOTC_enabled = False
IOTC_interval_between_reports_seconds       = 150.0 # For 20,000 messages/month the maximum interval is 133.92 seconds
IOTC_scopeID        = None
IOTC_deviceID       = None
IOTC_primaryKey     = None
IOTC_secondaryKey   = None

REST_enabled                                = False
REST_interval_between_reports_seconds       = 30.0
REST_Content_type                           = "application/json"
REST_URL                                    = "https://my.rayven.io:8082/api/main?uid=181829f4d6af889b418ababf3428c5fc0044&deviceid="

def InitializeConfiguration():
    global BMM150_minimum_SNR_for_speed_detection_dB
    global BMM150_quantile_for_noise_floor_estimator

    global ASYNC_MOTOR_number_of_poles_pairs
    global ASYNC_MOTOR_slip_factor_percentages

    global IOTC_enabled
    global IOTC_interval_between_reports_seconds
    global IOTC_scopeID
    global IOTC_deviceID
    global IOTC_primaryKey
    global IOTC_secondaryKey

    global REST_enabled
    global REST_interval_between_reports_seconds
    global REST_Content_type
    global REST_URL

    config = configparser.ConfigParser(allow_no_value=False)
    config.read(filenames="iCOMOX_GUI.ini")

    BMM150_quantile_for_noise_floor_estimator = config.getfloat(section="BMM150", option="BMM150_quantile_for_noise_floor_estimator", fallback=0.25)
    BMM150_minimum_SNR_for_speed_detection_dB = config.getfloat(section="BMM150", option="BMM150_minimum_SNR_for_speed_detection_dB", fallback=25)

    ASYNC_MOTOR_number_of_poles_pairs = config.getint(section="ASYNC_MOTOR", option="ASYNC_MOTOR_number_of_poles_pairs", fallback=2)
    ASYNC_MOTOR_slip_factor_percentages = config.getfloat(section="ASYNC_MOTOR", option="ASYNC_MOTOR_slip_factor_percentages", fallback=0.0)

    IOTC_enabled = config.getboolean(section="IOTC", option="IOTC_enabled", fallback=False)
    IOTC_interval_between_reports_seconds = config.getfloat(section="IOTC", option="IOTC_interval_between_reports_seconds", fallback=150.0)
    if IOTC_enabled:
        IOTC_scopeID = config.get(section="IOTC", option="IOTC_scopeID", fallback=None)
        IOTC_deviceID = config.get(section="IOTC", option="IOTC_deviceID", fallback="")
        IOTC_primaryKey = bytearray.fromhex(config.get(section="IOTC", option="IOTC_primaryKey", fallback=""))
        IOTC_secondaryKey = bytearray.fromhex(config.get(section="IOTC", option="IOTC_secondaryKey", fallback=""))
    else:
        IOTC_scopeID = ""
        IOTC_deviceID = ""
        IOTC_primaryKey = bytearray()
        IOTC_secondaryKey = bytearray()

    REST_enabled = config.getboolean(section="REST", option="REST_enabled", fallback=False)
    REST_interval_between_reports_seconds = config.getfloat(section="REST", option="REST_interval_between_reports_seconds", fallback=150.0)
    REST_Content_type = config.get(section="REST", option="REST_Content-type", fallback="application/json")
    REST_URL = config.get(section="REST", option="REST_URL", fallback="https://my.rayven.io:8082/api/main?uid=181829f4d6af889b418ababf3428c5fc0044&deviceid=")

    del config


from dataclasses import dataclass
from typing import List, Dict, Optional
from txp.common.protos import gateway_config_pb2 as gateway_config_proto
from txp.common.configuration.sampling_window import SamplingWindow
import dataclasses


@dataclass
class GatewayConfig:
    """The Gateway configuration class.

    Received configurations from devices management services should be converted
    to instances of this class.

    Args:
        gateway_id: The Gateway ID configured in the management interface.
        registry_id: The Registry ID to which the Gateway belongs.
        project_id: The GCP Project ID to which the Gateway belongs.
        cloud_region: The cloud region of the project in GCP.
        machines_ids: The List of MachineMetadata IDs monitored by the gateway.
    """

    gateway_id: str
    registry_id: str
    project_id: str
    cloud_region: str
    machines_ids: List[str]
    tenant_id: str

    def get_proto(self) -> gateway_config_proto.GatewayConfigProto:
        """Returns the Protobuff message for this GatewayConfig instance.

        Returns:
            GatewayConfigProto message.
        """

        def map_sampling_window_to_proto(sampling_window: SamplingWindow):
            proto = gateway_config_proto.SamplingWindowProto(
                observation_time=sampling_window.observation_time,
                sampling_time=sampling_window.sampling_time,
            )
            return proto

        config_proto = gateway_config_proto.GatewayConfigProto(
            gateway_id=self.gateway_id,
            registry_id=self.registry_id,
            project_id=self.project_id,
            cloud_region=self.cloud_region,
        )
        config_proto.machines_ids.extend(self.machines_ids)

        return config_proto

    @classmethod
    def build_from_dict(cls, d: Dict) -> Optional["GatewayConfig"]:
        known_fields = list(map(
            lambda field: field.name in d,
            dataclasses.fields(cls)
        ))

        # All the known fields are in the dict
        if not all(known_fields):
            return None

        # Remove unknown field from the dict
        known_fields = list(map(
            lambda field: field.name,
            dataclasses.fields(cls)
        ))

        invalid_keys = [key if key not in known_fields else None for key in d.keys()]

        new_dict = d.copy()
        for key in invalid_keys:
            if key:
                new_dict.pop(key)

        gateway_config = cls(**new_dict)
        return gateway_config

    @staticmethod
    def build_from_proto(
        proto: gateway_config_proto.GatewayConfigProto,
    ) -> "GatewayConfig":
        """Returns the class instance from the proto value.

        Args:
            proto: The GatewayConfigProto to build the class instance from.

        Raises:
            ValueError: if the received proto is not an instance of GatewayConfigProto.
        """
        if isinstance(proto, gateway_config_proto.GatewayConfigProto):
            gateway_id = proto.gateway_id
            registry_id = proto.registry_id
            project_id = proto.registry_id
            cloud_region = proto.cloud_region

            machines_metadata_ids = list(proto.machines_ids)

            return GatewayConfig(
                gateway_id,
                registry_id,
                project_id,
                cloud_region,
                machines_metadata_ids,
                ""  # Empty tenant-id because configuration protobuff are NOT being used.
            )

        raise ValueError(
            "Trying to build GatewayConfig from proto, but the proto instance is unknown"
        )

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON.

        The Dictionary will contain the values required to re-create the instance from the JSON
        object in a pythonic way:
            object = MyObject( **json.loads(json_string) )
        """
        return dict(
            gateway_id=self.gateway_id,
            registry_id=self.registry_id,
            project_id=self.project_id,
            cloud_region=self.cloud_region,
            machines_ids=self.machines_ids,
            tenant_id=self.tenant_id
        )

    def __eq__(self, other):
        """Equality operator impementation to compare two instances of
        GatewayConfig"""
        if not isinstance(self, other):
            return False

        return (
                self.gateway_id == other.gateway_id and
                self.registry_id == other.registry_id and
                self.project_id == other.project_id and
                self.cloud_region == other.cloud_region and
                self.machines_ids == other.machines_ids and
                self.tenant_id == other.tenant_id
        )



cESCAPE_CHAR        = 255

cPREV_ESC_CHAR_00   = 1
cPREV_ESC_CHAR_08   = 2
cPREV_ESC_CHAR_1A   = 3
cPREV_ESC_CHAR_1B   = 4
cPREV_ESC_CHAR_FF   = 5

# [DecodeOk, prev_escaped_char, data_out] = decode(data_in, prev_escaped_char)
# IN PARAMETERS:
# data_in - bytearray with the encoded data,
# prev_escaped_char - if the last data_in in previous call to decode() has ended with the cESCAPE_CHAR
# OUT PARAMETERS:
# DecodeOk - is the decoding succeeded
# prev_escaped_char - did the last byte in data_in is cESCAPE_CHAR (should be transferred to the next call to decode())
# data_out - bytearray which contains the decoded array
def decode(data_in, prev_escaped_char):
    #   0x00 -> 0xFF 0x01
    #   0x08 -> 0xFF 0x02
    #   0x1A -> 0xFF 0x03
    #   0x1B -> 0xFF 0x04
    #   0xFF -> 0xFF 0x05
    data_out = bytearray()
    for i in range(0, len(data_in)):
        if prev_escaped_char:
            if data_in[i] == cPREV_ESC_CHAR_00:
                data_out += b"\x00"
            elif data_in[i] == cPREV_ESC_CHAR_08:
                data_out += b"\x08"
            elif data_in[i] == cPREV_ESC_CHAR_1A:
                data_out += b"\x1A"
            elif data_in[i] == cPREV_ESC_CHAR_1B:
                data_out += b"\x1B"
            elif data_in[i] == cPREV_ESC_CHAR_FF:
                data_out += b"\xFF"
            else:
                return False, False, data_out  # [ decode fail = False, prev_escaped_char = False ]
            prev_escaped_char = False
        else:
            if data_in[i] == cESCAPE_CHAR:
                prev_escaped_char = True
            elif (data_in[i] == 0) or (data_in[i] == 8) or (data_in[i] == 26) or (data_in[i] == 27):
                return False, False, data_out
            else:
                data_out += bytearray([data_in[i]])
    return True, prev_escaped_char, data_out

# [prev_escaped_char, data_out] = encode(data_in, prev_escaped_char)
# IN PARAMETERS:
# data_in - bytearray with the non-encoded data,
# OUT PARAMETERS:
# data_out - bytearray which contains the encoded array
def encode(data_in):
    #   0x00 -> 0xFF 0x01
    #   0x08 -> 0xFF 0x02
    #   0x1A -> 0xFF 0x03
    #   0x1B -> 0xFF 0x04
    #   0xFF -> 0xFF 0x05
    data_out = bytearray()
    for i in range(0, len(data_in)):
        if data_in[i] == 0:
            data_out += b"\xFF\x01"
        elif data_in[i] == 8:
            data_out += b"\xFF\x02"
        elif data_in[i] == 26:
            data_out += b"\xFF\x03"
        elif data_in[i] == 27:
            data_out += b"\xFF\x04"
        elif data_in[i] == 255:
            data_out += b"\xFF\x05"
        else:
            data_out += bytearray([data_in[i]])
    return data_out



"""
    This module provides functions and classes that helps with
    programmatic interaction with the BigQuery service.
"""
from google.cloud import bigquery
from datetime import datetime
from txp.common.config import settings
import PIL
from PIL import Image
import io
import logging
import numpy as np
from txp.cloud.common import utils as cu
import time
from scipy.fft import fft

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)


def get_image_from_raw_data(data, lower_quality: bool = True):
    image = [int(x) for x in data[0]["values"]]
    image = np.frombuffer(bytes(image), dtype=np.uint8)
    image_b = bytes(image)
    try:
        image_pil: Image = Image.open(io.BytesIO(image_b))

        if not lower_quality:
            return image_pil

        with io.BytesIO() as output:
            image_pil.save(output, format="JPEG", quality=30, optimize=True)
            new_image = Image.open(io.BytesIO(output.getvalue()))

        return new_image

    except PIL.UnidentifiedImageError as e:
        log.info(f"Found a corrupted image: {e}")
        return None
    except Exception as e:
        log.info(f"Found a corrupted image: {e}")
        return None


def get_authenticated_client(service_account_credentials) -> bigquery.Client:
    """Returns a bigquery.Client, authenticated with the provided credentials.

    Args:
        service_account_credentials: A service_account.Credentials object with the
            appropriate permissions to use Bigquery.
    """
    return bigquery.Client(credentials=service_account_credentials)


def get_all_records_within_interval(
        tenant_id: str,
        table_name: str,
        edge_logical_id: str,
        perception_name: str,
        start_datetime: datetime,
        end_datetime: datetime,
        order_key: str,
        client: bigquery.Client
):
    """Gets all records within interval [start_datetime, end_datetime]
        Args:
            tenant_id: id of tenant
            table_name: bigquery table name
            edge_logical_id: logical id for edge
            perception_name: perception name for given edge
            start_datetime: interval lower bound
            end_datetime: interval upper bound
            order_key: resulting dataframe will be ordered by this key
            client: bigquery client
        Returns:
            List of signals, each element of the list is a dictionary
    """

    start_time = int(start_datetime.timestamp() * 1e9)
    end_time = int(end_datetime.timestamp() * 1e9)
    start_partition_timestamp = cu.get_partition_utc_date(start_time)
    end_partition_timestamp = cu.get_partition_utc_date(end_time + 3.6e12)

    select_query = f"""
           SELECT * FROM `{table_name}` 
           WHERE tenant_id = "{tenant_id}"
                 AND edge_logical_id = "{edge_logical_id}" 
                 AND perception_name = "{perception_name}"
                 AND observation_timestamp >= {start_time}
                 AND observation_timestamp < {end_time}
                 AND partition_timestamp >= "{start_partition_timestamp}"
                 AND partition_timestamp < "{end_partition_timestamp}"
           ORDER BY {order_key} DESC;
           """

    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds for perception: {perception_name}")

    return df


def get_all_task_predictions_within_interval(
        tenant_id: str,
        table_name: str,
        asset_id: str,
        start_datetime: datetime,
        end_datetime: datetime,
        client: bigquery.Client
):
    """Gets all records within interval [start_datetime, end_datetime]
        Args:
            tenant_id: id of tenant
            table_name: bigquery table name
            asset_id: logical id for asset
            start_datetime: interval lower bound
            end_datetime: interval upper bound
            client: bigquery client
        Returns:
            List of predictions (events or states), df containing all events
    """

    start_time = int(start_datetime.timestamp() * 1e9)
    end_time = int(end_datetime.timestamp() * 1e9)
    start_partition_timestamp = cu.get_partition_utc_date(start_time)
    end_partition_timestamp = cu.get_partition_utc_date(end_time + 3.6e12)

    select_query = f"""
           SELECT * FROM `{table_name}` 
           WHERE tenant_id = "{tenant_id}"
                 AND asset_id = "{asset_id}" 
                 AND partition_timestamp >= "{start_partition_timestamp}"
                 AND partition_timestamp < "{end_partition_timestamp}"
           ORDER BY observation_timestamp ASC;
           """

    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds")
    return df


def get_events_for_state(
        tenant_id,
        table_name: str,
        asset_id: str,
        events_id: list,
        days: int,
        client: bigquery.Client
):
    offset = days * 86400
    end = time.time_ns() // 1e9
    start = end - offset
    start_date = str(datetime.utcfromtimestamp(start))
    end_date = str(datetime.utcfromtimestamp(end))

    select_query = f""" 
       SELECT * FROM `{table_name}` 
       WHERE tenant_id = "{tenant_id}" 
             AND asset_id = "{asset_id}"
             AND event_id IN UNNEST({events_id})
             AND partition_timestamp <= "{end_date}"
             AND partition_timestamp >= "{start_date}"
       """

    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds")
    return df.to_dict('records')


def get_all_signals_within_interval(
        tenant_id: str,
        table_name: str,
        edge_logical_id: str,
        perception_name: str,
        start_datetime: datetime,
        end_datetime: datetime,
        client: bigquery.Client
):
    """Gets all time signals within interval [start_datetime, end_datetime]
        Args:
            tenant_id: id of tenant
            table_name: bigquery table name
            edge_logical_id: logical id for edge
            perception_name: perception name for given edge
            start_datetime: interval lower bound
            end_datetime: interval upper bound
            client: bigquery client
        Returns:
            List of signals, each element of the list is a dictionary
    """

    df = get_all_records_within_interval(tenant_id, table_name, edge_logical_id, perception_name, start_datetime,
                                         end_datetime, "signal_timestamp", client)

    grouped_df = df.groupby("signal_timestamp")

    res = []

    for signal_timestamp, signal_df in grouped_df:
        all_chunks = signal_df.to_dict('records')
        signal = cu.merge_signal_chunks(all_chunks)
        res.append(signal)
    return res


def get_last_task_prediction_for_asset(
        tenant_id,
        table_name: str,
        asset_id: str,
        days: int,
        client: bigquery.Client
):
    offset = days * 86400
    end = time.time_ns() // 1e9
    start = end - offset
    start_date = str(datetime.utcfromtimestamp(start))
    end_date = str(datetime.utcfromtimestamp(end))
    select_query = f""" 
       SELECT * FROM `{table_name}` 
       WHERE tenant_id = "{tenant_id}" 
             AND asset_id = "{asset_id}"
             AND partition_timestamp <= "{end_date}"
             AND partition_timestamp >= "{start_date}"
       ORDER BY observation_timestamp DESC;
       """

    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds")

    if not len(df):
        return None

    return df.iloc[0].to_dict()


def get_fft_from_raw_data(data):
    fft_per_dimensions = []
    for index, dimension in enumerate(data):
        fft_per_dimensions.append(fft(dimension["values"]))
    return np.array(fft_per_dimensions)


def get_last_signal_from_bigquery(tenant_id, table_name, edge_logical_id, perception_name, start_time, end_time,
                                  client):
    """Gets the latest signal  in the interval [start_time, end_time) that is store in bigquery given an
    edge_logical_id and a perception_name

    Args:
        tenant_id: id of tenant.
        perception_name: perception name of the signal.
        edge_logical_id: edge that generated the signal.
        start_time: left bound of time interval
        end_time: right bound of time interval
        table_name: table name
        client: BigQuery client
    """
    signals = get_all_signals_within_interval(tenant_id, table_name, edge_logical_id, perception_name, start_time,
                                              end_time, client)
    if not len(signals):
        return None
    return max(signals, key=lambda signal: signal["signal_timestamp"])


def get_all_signals_for_asset(tenant_id, table_name, edge_logical_ids, dataset_versions,
                              start_datetime, end_datetime, client):
    start_time = int(start_datetime.timestamp() * 1e9)
    end_time = int(end_datetime.timestamp() * 1e9)
    start_partition_timestamp = cu.get_partition_utc_date(start_time)
    end_partition_timestamp = cu.get_partition_utc_date(end_time + 3.6e12)

    tags_query = ""
    for i, version in enumerate(dataset_versions):
        tags_query += f'"{version}" IN UNNEST(dataset_versions) '
        if i + 1 < len(dataset_versions):
            tags_query += "OR "
    select_query = f"""
           SELECT * FROM `{table_name}` 
           WHERE tenant_id = "{tenant_id}"
                 AND edge_logical_id IN UNNEST({edge_logical_ids})
                 AND ({tags_query})
                 AND observation_timestamp >= {start_time}
                 AND observation_timestamp <= {end_time}
                 AND partition_timestamp >= "{start_partition_timestamp}"
                 AND partition_timestamp < "{end_partition_timestamp}";
           """

    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds, with {len(df)} registers")
    log.debug(f"{select_query}")
    return df


def annotate_signal(tenant_id, table_name, edge_logical_id, observation_timestamp, label,
                    dataset_versions, client):
    partition_timestamp = cu.get_partition_utc_date(observation_timestamp)

    label_set_str = f"""label = '''{label}''', """ if table_name.split(".")[1] == "time" else ""

    select_query = f"""
       UPDATE `{table_name}`
       SET {label_set_str}
           dataset_versions = dataset_versions || ["{dataset_versions}"]  
       WHERE tenant_id = "{tenant_id}"
             AND edge_logical_id = "{edge_logical_id}"
             AND observation_timestamp = {observation_timestamp}
             AND partition_timestamp = "{partition_timestamp}";
    """
    start = time.time()
    res = client.query(select_query).result()
    end = time.time()
    log.info(f"This query took: {end - start} seconds")
    log.info(f"Table name: {table_name}")
    log.info(f"{select_query}")

    return res


def get_sampling_window_from_bigquery(table_id, edge_logical_id, perception_name, observation_timestamp, tenant_id,
                                      client):
    partition_timestamp = str(cu.get_partition_utc_date(observation_timestamp))
    select_query = f"""
               SELECT * FROM `{table_id}` 
               WHERE tenant_id = "{tenant_id}"
                     AND edge_logical_id = "{edge_logical_id}" 
                     AND observation_timestamp = {observation_timestamp}
                     AND perception_name = "{perception_name}"
                     AND partition_timestamp = "{partition_timestamp}"
               """
    start = time.time()
    df = (client.query(select_query).result().to_dataframe())
    end = time.time()
    log.info(f"This query took: {end - start} seconds")
    return df.to_dict('records')



import time
from typing import Dict
from txp.common.protos import gateway_config_pb2 as gateway_config_proto


class SamplingWindow:
    """Configuration values for the sampling cadence of Edge devices.
    The sampling cadence defines how to handle data collected by edge
    devices according to time.
    The sampling cadence contemplates the following configuration parameters:
        Observation time:
            time where a full sampling is collected, which can include device collected
            data and human-in-the-loop generated data.
        Sampling time:
            the period of time within an observation time where device collected data should be considered.
        Dead time:
            the period of time within an observation time where edge devices aren't collecting data,
            or it is being discarded.

    A SamplingWindow refers to the Machine identifiers of the machine(s) that are configured with
    that sampling window.
    """

    def __init__(self, observation_time: int, sampling_time: int, observation_timestamp=None):
        """
        Args:
            observation_time (int): The Observation time value.
            sampling_time (int): The Sampling time value.
            machines_ids(str): The MachineMetadata IDs associated with this
                sampling window.
        """
        self._observation_timestamp = observation_timestamp
        self._observation_time: int = int(observation_time)
        self._sampling_time: int = int(sampling_time)
        self._dead_time: int = abs(observation_time - sampling_time)

        # Set at runtime for the sampling windows task context
        self._gateway_task_id: int = 0
        self._sampling_window_index: int = 0
        self._number_of_sampling_windows: int = 0

    @property
    def observation_time(self) -> int:
        return self._observation_time

    @property
    def gateway_task_id(self) -> int:
        return self._gateway_task_id

    @property
    def sampling_window_index(self) -> int:
        return self._sampling_window_index

    @property
    def number_of_sampling_windows(self) -> int:
        return self._number_of_sampling_windows

    @gateway_task_id.setter
    def gateway_task_id(self, task_id: int) -> None:
        self._gateway_task_id = task_id

    @sampling_window_index.setter
    def sampling_window_index(self, index: int) -> None:
        self._sampling_window_index = index

    @number_of_sampling_windows.setter
    def number_of_sampling_windows(self, num_sampling_windows: int) -> None:
        self._number_of_sampling_windows = num_sampling_windows

    @property
    def sampling_time(self) -> int:
        return self._sampling_time

    @property
    def observation_timestamp(self) -> int:
        return self._observation_timestamp

    @observation_timestamp.setter
    def observation_timestamp(self, timestamp):
        self._observation_timestamp = timestamp

    @property
    def dead_time(self) -> int:
        return self._dead_time

    @staticmethod
    def build_from_proto(
        proto: gateway_config_proto.SamplingWindowProto,
    ) -> "SamplingWindow":
        """Returns the class instance from the proto value.

        Args:
            proto: The SamplingWindowProto to build the class instance from.

        Raises:
            ValueError: if the received proto is not an instance of SamplingWindowProto.
        """
        if isinstance(proto, gateway_config_proto.SamplingWindowProto):
            sw = SamplingWindow(
                proto.observation_time, proto.sampling_time, proto.observation_timestamp,
            )
            sw.number_of_sampling_windows = proto.number_of_sampling_windows
            sw.sampling_window_index = proto.sampling_window_index
            sw.gateway_task_id = proto.gateway_task_id
            return sw

        raise ValueError(
            "Trying to build SamplingWindow from proto, but the proto instance is unknown"
        )

    def get_proto(self) -> gateway_config_proto.SamplingWindowProto:
        """Return the SamplingWindowProto instance for this SamplingWindow"""
        proto = gateway_config_proto.SamplingWindowProto(
            observation_time=self.observation_time,
            sampling_time=self.sampling_time,
            observation_timestamp=self.observation_timestamp,
            gateway_task_id=self.gateway_task_id,
            sampling_window_index=self.sampling_window_index,
            number_of_sampling_windows=self.number_of_sampling_windows
        )
        return proto

    def reprJSON(self) -> Dict:
        """Returns a Dict easily serializable to JSON"""
        return dict(observation_time=self.observation_time,
                    sampling_time=self.sampling_time,
                    observation_timestamp=self.observation_timestamp,
                    gateway_task_id=self.gateway_task_id,
                    sampling_window_index=self.sampling_window_index,
                    number_of_sampling_windows=self.number_of_sampling_windows
                    )



"""
This module implements a generic USB video driver based on OpenCV library.
"""
# ============================ imports =========================================

import threading
import cv2
from typing import List
from txp.devices.drivers.driver import DriverState, StateReport
from txp.devices.drivers.usb.usb_base_driver import UsbDriver
from txp.common.edge import EdgeDescriptor

from txp.devices.package_queue import PackageQueueProxy
from txp.common.edge.perception_dimensions import ImageSignal
from txp.common.config import settings
import queue
import logging

log = logging.getLogger(__name__)
log.setLevel(settings.txp.general_log_level)

# ==============================================================================
# USB Camera Driver Implementation.
# ==============================================================================


class UsbVideoDriver(UsbDriver):
    """Implementation of the USB video camera driver.
    """

    def __init__(
            self,
            edge_descriptor: EdgeDescriptor,
            gateway_packages_queue: PackageQueueProxy,
            on_driver_state_changed: callable,
            host_clock_notification_queue: queue.Queue
    ):
        """
        Note: This Driver receives a USB port address, which is added and passed down by the Gateway operative
        configuration. The USB port address is discovered during the production stage.

        Args
            edge_descriptor: The Edge Descriptor for this Driver.
            gateway_packages_queue: The Gateway Package Queue to put produced packages.
            on_driver_state_changed: Function used to notify driver state changes.
        """
        UsbDriver.__init__(self, edge_descriptor, gateway_packages_queue,
                           on_driver_state_changed, host_clock_notification_queue)
        self.camera_handler = None

    def is_virtual(cls) -> bool:
        return False

    def observation_pulse(self) -> int:
        return 5

    def connect_device(self):
        """Connects to the USB camera with the corresponding physical id.
        """
        log.info(f"Connecting USB camera {self.logical_id}")
        if not self.camera_handler:
            self.camera_handler = cv2.VideoCapture(self._physical_id)

        if not self.camera_handler.isOpened():
            log.error(f"Error connecting to the USB physical_id: {self._physical_id}:")
        else:
            self.on_driver_state_changed(
                self.logical_id, DriverState.CONNECTED
            )

    def disconnect_device(self):
        """Disconnects the USB camera and release the handler.
        """
        if self.camera_handler:
            self.camera_handler.release()
            log.info(f"USB camera {self.logical_id} was disconnected")
        self.on_driver_state_changed(
            self.logical_id, DriverState.DISCONNECTED
        )

    def _start_sampling_collect(
            self
    ) -> List["Signals"]:
        """Collects the signals from the camera and generates the packages for the gateway.

        This method should collect samples that belong to the defined Sampling Time.

        Args:
            sampling_time_finished: event to check when the sampling time has finished.
        Returns:
            A list of ImageSignal collected by the USB camera.
        """
        signals = []

        image = self.get_image()

        signal = ImageSignal.build_signal_from_lists(
            [image],
            (1, len(image)),
            0,
            0
        )
        signals.append(signal)

        return signals

    def _stop_sampling_collect(self):
        """This method will stop the sampling collection in the driver implementation.
        """
        pass

    def get_image(self):
        ret, image = self.camera_handler.read()
        log.info(f"Image taken from {self.logical_id}")
        if not ret:
            log.error("Error reading image")

        ret, image_jpeg = cv2.imencode('.jpeg', image)
        if not ret:
            log.error("Error encoding JPEG image")

        return list(image_jpeg)

    def get_video(self):
        # TODO: Not Supported yet
        pass



import enum
from typing import List


class AppProfile(enum.Enum):
    def __new__(cls, *values):
        obj = object.__new__(cls)
        obj._value_ = values[0]
        for other_value in values[1:]:
            cls._value2member_map_[other_value] = obj
        obj._all_values = values
        obj.display_value = values[1]
        return obj

    """This enumeration defines the valid Application Profiles supported."""
    RESUMEN = 0, "Resumen"
    ANALYTICS = 1, "Análisis"
    CONFIGURATION = 2, "Configuración"

    def get_display_value(self) -> str:
        return self.display_value

    @classmethod
    def get_display_values(cls) -> List[str]:
        """Returns the List of display value strings for the profiles"""
        r = []
        for profile in cls:
            r.append(profile.display_value)
        return r



'''
Base class for ApiConnectors

The base ApiConnector class contains common methods for managing the notification queue. 
'''

import logging
from queue import Empty
from queue import Queue

from SmartMeshSDK import ApiException
DEFAULT_Q_SIZE = 1000

# Log initialization 
class NullHandler(logging.Handler):
    def emit(self, record):
        pass
log = logging.getLogger('ApiConnector')
log.setLevel(logging.INFO)
log.addHandler(NullHandler())

def logDump(buf, msg = None, level=logging.DEBUG):
    '''
    \brief Print dump for binary object to trace file.
    '''
    
    if log.getEffectiveLevel() > level :
        return
    if msg :
        log.log(level, msg)
    addr = 0
    step = 20
    dump = " ".join(["{0:02x}".format(ord(c)) for c in buf])
    for i in xrange(0, len(dump), 3 * step) :
        log.log(level, "    {0:3} : {1}".format(addr, dump[i : i + 3 * step]))
        addr += step
        

class ApiConnector(object):
    '''
    \ingroup ApiConnector
    
    \brief Base class for all connector objects.
    '''
    
    def __init__(self, maxQSize = DEFAULT_Q_SIZE):
        self.maxQSize = maxQSize
        self.queue = NotifQueue(self.maxQSize)
        self.isConnected = False
        self.isExceptionRaise = False
        self.pendingNotification = None
        self.disconnectReason = ''
        self.traceFile = None
                
    def connect(self) :
        self.queue.clear()
        self.isConnected = True
        self.isExceptionRaise = False
        self.pendingNotification = None
        
    def disconnect(self, reason):
        if not self.isConnected :
            return
        self.disconnectReason = reason
        self.isConnected = False
        self.putDisconnectNotification(reason)
        
    def send(self, cmdName, params):
        raise NotImplementedError("ApiConnector.send is not implemented")

    def getNotificationInternal(self, timeoutSec=-1):
        '''
        \brief get notification from queue
        
        \param timeoutSec timeout for waiting if queue is empty.
               <0 wait infinity (blocked), >=0 wait up to 'timeout' seconds
         
        \exception ConnectionError disconnected from device
        \exception QueueError reading from empty 'offline' queue
        \returns   Notification object or None if queue is empty.
        '''
        
        if self.pendingNotification :
            res, self.pendingNotification = self.pendingNotification, None
            return res
        
        if not self.isConnected :
            self.oneTimeRaiseDisconnectException(None)
            timeoutSec = 0   # for 'offline' queue use get without timeout 
            
        res = self.queue.get(timeoutSec)
        if not self.isConnected :
            self.oneTimeRaiseDisconnectException(res)
            
        if not self.isConnected and not res :
            raise ApiException.QueueError() # Send exception: Reading from empty queue
        return res
    
    def oneTimeRaiseDisconnectException(self, notif):
        '''
        \brief raise exception only one time for one session and save current
               notification.
        '''
        
        if not self.isExceptionRaise :  # Send Disconnect Exception (only one time)
            self.isExceptionRaise = True
            self.pendingNotification = notif
            raise ApiException.ConnectionError(self.disconnectReason)

    def putNotification(self, item):
        '''
        \brief Put notification to queue
         
        Insert notification to queue. If queue is full raise ConnectionError exception
       
        \param item notification to insert
       
        \exception ConnectionError
        '''
        
        if not self.isConnected :
            raise ApiException.ConnectionError("Disconnected")
        if self.queue.qsize() < self.maxQSize :
            self.queue.put(item)
        else :
            raise ApiException.ConnectionError("Queue overflowed")

    def putDisconnectNotification(self, reason):
        '''
        \brief Put Disconnect notification to queue
       
        \param reason reason for disconnection
        '''
        # used by HartMgr where notifications are separate from control messages
        self.queue.putDisconnectNotification(reason)
        
     
class NotifQueue(Queue):
    class _DisconnectNotification:
        '''
        \brief Special internal notification - connection is broken
        '''
        def __init__(self, reason):
            self.reason = reason
        
    def __init__(self, maxSize):
        self.maxQSize = maxSize
        Queue.__init__(self)
    
    def get(self, timeout = -1):
        '''
        \brief Get notification from queue
        
        \param timeout timeout for waiting if queue is empty.
               <0 wait forever (blocked), >=0 wait up to 'timeout' seconds
        
        \returns   Notification object or None if queue is empty.
        '''
        
        try :
            if timeout < 0 :
                notif = Queue.get(self, True)
            elif timeout == 0 :
                notif = Queue.get(self, False)
            else:
                notif = Queue.get(self, True, timeout)
            if isinstance(notif, NotifQueue._DisconnectNotification) :  
                notif = Queue.get(self, False)  # disconnect notification is used for kick 'get' method  
            return notif
        except Empty:
            return None 
               
    def putDisconnectNotification(self, reason):
        '''
        \brief Put Disconnect notification to queue
       
        \param reason reason for disconnection
        '''
        self.put(NotifQueue._DisconnectNotification(reason))
        
    def clear(self) :
        while self.get(0) : pass
        



'''API Definition for HART Manager XML API'''

from SmartMeshSDK.ApiDefinition import ApiDefinition

import xmlutils
import re

# Add a log handler for the HART Manager

import logging
class NullHandler(logging.Handler):
    def emit(self, record):
        pass

log = logging.getLogger('HartManager')
log.setLevel(logging.INFO)
log.addHandler(NullHandler())

##
# \ingroup ApiDefinition
#

class HartMgrDefinition(ApiDefinition.ApiDefinition):
    '''
    \brief API definition for the HART manager.
   
    \note This class inherits from ApiDefinition. It redefines the attributes of
          its parent class, but inherits the methods.
    '''

    FIELDS = 'unnamed'
    
    STRING    = ApiDefinition.FieldFormats.STRING
    BOOL      = ApiDefinition.FieldFormats.BOOL
    INT       = ApiDefinition.FieldFormats.INT
    INTS      = ApiDefinition.FieldFormats.INTS
    FLOAT     = ApiDefinition.FieldFormats.FLOAT
    HEXDATA   = ApiDefinition.FieldFormats.HEXDATA
    RC        = ApiDefinition.ApiDefinition.RC
    SUBID1    = ApiDefinition.ApiDefinition.SUBID1
    LIST      = 'list'

    # Enumerations
    # fieldOptions is a list of [ value, description, ?? ]
    fieldOptions = {
        RC : [
            [0x0,                      'OK',                    ''],
            # ...
        ],
        'bool' : [
            ['true',                   'true',                  ''],
            ['false',                  'false',                 ''],
        ],
        'appDomain' : [
            ['maintenance',            'maintenance',           ''],
        ],
        'packetPriority' : [
            ['low',                    'low',                   ''],
            ['high',                   'high',                  ''],
        ],
        'pipeDirection' : [
            ['UniUp',                  'upstream',              ''],
            ['UniDown',                'downstream',            ''],
            ['Bi',                     'bidirectional',         ''],
        ],
        'moteState' : [
            ['Idle',                   'idle',                  ''],
            ['Lost',                   'lost',                  ''],
            ['Joining',                'joining',               ''],
            ['Operational',            'operational',           ''],
            ['Disconnecting',          'disconnecting',         ''],
        ],
        'pathDirection' : [
            ['all',                    'all',                   ''],
            ['upstream',               'upstream',              ''],
            ['downstream',             'downstream',            ''],
            ['unused',                 'unused',                ''],
        ],
        'bandwidthProfile' : [
            ['Manual',                 'manual profile',        ''],
            ['P1',                     'normal profile',        ''],
            ['P2',                     'low-power profile',     ''],
        ], 
        'securityMode' : [
            ['acceptACL',              'Accept ACL',             ''],
            ['acceptCommonJoinKey',    'Accept common join key', '']
        ],
        'userPrivilege' : [
            ['viewer',                 'viewer',                ''],
            ['user',                   'user',                  ''],
            ['superuser',              'superuser',             ''],
        ],
        'onOff' : [
            ['on',                     'on',                    ''],
            ['off',                    'off',                   ''],
        ],
        'resetObject' : [
            ['network',                'network',               ''],
            ['system',                 'system',                ''],
            ['stat',                   'statistics',            ''],
            ['eventLog',               'eventLog',              ''],
        ],
        'resetMote' : [
            ['mote',                   'mote',                  ''],
        ],
        'statPeriod': [
            ['current',                'current',               ''],
            ['lifetime',               'lifetime',              ''],
            ['short',                  'short',                 ''],
            ['long',                   'long',                  ''],
        ],
        'advertisingStatus': [
            ['on',                     'on',                    ''],
            ['off',                    'off',                   ''],
            ['pending',                'pending',               ''],
        ],
        'pipeStatus': [
            ['off',                    'off',                     ''],
            ['pending',                'Pipe activation pending', ''],
            ['on_bi',                  'Bidirection pipe on',     ''],
            ['on_up',                  'Upstream pipe on',        ''],
            ['on_down',                'Downstream pipe on',      ''],
        ],
        'locationTag': [
            ['supported',              'supported',               ''],
            ['not supported',          'not supported',           ''],
        ],
        'redundancyMode': [
            ['standalone',             'standalone',              ''],
            ['transToMaster',          'Transitioning to master', ''],
            ['transToSlave',           'Transitioning to slave',  ''],
            ['master',                 'master',                  ''],
            ['slave',                  'slave',                   ''],
            ['failed',                 'Manager failed',          ''],
        ],
        'redundancyPeerStatus': [
            ['unknown',                'unknown',                 ''],
            ['connected',              'connected',               ''],
            ['synchronized',           'synchronized',            ''],
        ],
        'channelType': [
            ['cli',                    'Manager CLI',             ''],
            ['config',                 'API control',             ''],
            ['notif',                  'API notifications',       ''],
        ],
        'alarmType': [
            ['moteDown',               'Mote down alarm',         ''],
            ['slaReliability',         'SLA Reliability',         ''],
            ['slaLatency',             'SLA Latency',             ''],
            ['slaStability',           'SLA Stability',           ''],
            ['maxMotesReached',        'Maximum number of motes reached',  ''],
            ['bbLatencyWarn',          'Backbone latency warning',''],
        ],
    }


    # ----------------------------------------------------------------------
    # Notifications
    
    eventNotifications = [
        {
            'id'         : 'sysConnect',
            'name'       : 'UserConnect',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['channel',             STRING,   16,  'channelType'],
                    ['ipAddr',              STRING,   16,  None],
                    ['userName',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysDisconnect',
            'name'       : 'UserDisconnect',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['channel',             STRING,   16,  'channelType'],
                    ['ipAddr',              STRING,   16,  None],
                    ['userName',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualMoteReset',
            'name'       : 'ManualMoteReset',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualMoteDelete',
            'name'       : 'ManualMoteDelete',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualMoteDecommission',
            'name'       : 'ManualMoteDecommission',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualNetReset',
            'name'       : 'ManualNetReset',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualDccReset',
            'name'       : 'ManualDccReset',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysManualStatReset',
            'name'       : 'ManualStatReset',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'sysConfigChange',
            'name'       : 'ConfigChange',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['userName',            STRING,   32,  None],
                    ['objectType',          STRING,   32,  None], # TODO: enum
                    ['objectId',            STRING,   32,  None],
                ],
            },
        },

        {
            'id'         : 'sysBootUp',
            'name'       : 'BootUp',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    # None
                ],
            },
        },
        {
            'id'         : 'netReset',
            'name'       : 'NetworkReset',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    # None
                ],
            },
        },
        {
            'id'         : 'sysCmdFinish',
            'name'       : 'CommandFinished',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['callbackId',          INT,      4,   None],
                    ['objectType',          STRING,   32,  None], # TODO: enum
                    ['macAddr',             STRING,   32,  None],
                    ['resultCode',          INT,      4,   None],
                ],
            },
        },
        {
            'id'         : 'netPacketSent',
            'name'       : 'PacketSent',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['callbackId',          INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netMoteJoin',
            'name'       : 'MoteJoin',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                    ['userData',            STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteLive',
            'name'       : 'MoteLive',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteQuarantine',
            'name'       : 'MoteQuarantine',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteJoinQuarantine',
            'name'       : 'MoteJoinQuarantine',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                    ['userData',            STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteUnknown',
            'name'       : 'MoteUnknown',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteDisconnect',
            'name'       : 'MoteDisconnect',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteJoinFailure',
            'name'       : 'MoteJoinFailure',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['macAddr',             STRING,   32,  None],
                    ['reason',              STRING,   64,  None], # TODO: length
                ],
            },
        },
        {
            'id'         : 'netMoteInvalidMIC',
            'name'       : 'InvalidMIC',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPathCreate',
            'name'       : 'PathCreate',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   32,  None],
                    ['moteBMac',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPathDelete',
            'name'       : 'PathDelete',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   32,  None],
                    ['moteBMac',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPathActivate',
            'name'       : 'PathActivate',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   32,  None],
                    ['moteBMac',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPathDeactivate',
            'name'       : 'PathDeactivate',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   32,  None],
                    ['moteBMac',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPathAlert',
            'name'       : 'PathAlert',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   32,  None],
                    ['moteBMac',            STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netPipeOn',
            'name'       : 'PipeOn',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['macAddr',             STRING,   32,  None],
                    ['allocatedPipePkPeriod', INT,    4,   None],
                ],
            },
        },
        {
            'id'         : 'netPipeOff',
            'name'       : 'PipeOff',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['macAddr',             STRING,   32,  None],
                ],
            },
        },
        {
            'id'         : 'netServiceDenied',
            'name'       : 'ServiceDenied',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['serviceId',          INT,      4,   None],
                    ['requestingMacAddr',  STRING,   32,  None],
                    ['peerMacAddr',        STRING,   32,  None],
                    ['appDomain',          STRING,   32,  'appDomain'],
                    ['isSource',           BOOL,     1,   None],
                    ['isSink',             BOOL,     1,   None],
                    ['isIntermittent',     BOOL,     1,   None],
                    ['period',             INT,      4,   None],
                ],
            },
        },
        {
            'id'         : 'netPingReply',
            'name'       : 'PingReply',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['macAddr',             STRING,   32,  None],
                    ['callbackId',          INT,      4,   None],
                    ['latency',             INT,      4,   None],
                    ['temperature',         FLOAT,    8,   None],
                    ['voltage',             FLOAT,    8,   None],
                    ['hopCount',            INT,      4,   None],
                ],
            },
        },
        {
            'id'         : 'netTransportTimeout',
            'name'       : 'TransportTimeout',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['srcMacAddr',          STRING,   32,  None],
                    ['destMacAddr',         STRING,   32,  None],
                    ['timeoutType',         STRING,   32,  None], # TODO: timeout type
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },

        # TODO: redundancy events: sysRdntModeChange, sysRdntPeerStatusChange
        # TODO: alarm open and close have sub-events
    ]

    measurementNotifications = [
        {
            'id'         : 'location',
            'name'       : 'Location',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['ver',                 INT,      1,   None],
                    ['asn',                 INT,      8,   None],
                    ['src',                 STRING,   32,  None],
                    ['dest',                STRING,   32,  None],
                    ['payload',             HEXDATA, None, None],
                ],
            },
        },
    ]
    
    notifications = [
        {
            'id'         : 'event',
            'name'       : 'event',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['timeStamp',           INT,      8,   None],
                    ['eventId',             INT,      4,   None],
                    [SUBID1,                INT,      1,   None],
                ]
            },
            'subCommands': eventNotifications,
            'deserializer': 'parse_eventNotif',
        },
        {
            'id'         : 'data',
            'name'       : 'data',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   32,  None],
                    ['time',                INT,      8,   None],
                    ['payload',             HEXDATA, None, None],
                    ['payloadType',         INT,      1,   None],
                    ['isReliable',          BOOL,     1,   None],
                    ['isRequest',           BOOL,     1,   None],
                    ['isBroadcast',         BOOL,     1,   None],
                    ['callbackId',          INT,      4,   None],
                    # counter field added in 4.1.0.2
                    ['counter',             INT,      4,   None],
                ]
            },
            'deserializer': 'parse_dataNotif',
        },
        {
            'id'         : 'measurement',
            'name'       : 'measurement',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    [SUBID1,                INT,      1,   None],
                ]
            },
            'subCommands': measurementNotifications,
        },
        {
            'id'         : 'cli',
            'name'       : 'cli',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['time',                INT,      8,   None],
                    ['message',             STRING,   128, None],
                ]
            },
        },
        {
            'id'         : 'log',
            'name'       : 'log',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['time',                INT,      8,   None],
                    ['severity',            STRING,   16,  None],
                    ['message',             STRING,   128, None],
                ]
            },
        },
        {
            'id'         : 'stdMoteReport',
            'name'       : 'stdMoteReport',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['time',     INT,      8,   None],
                    ['macAddr',  STRING,   16,  None],
                    ['payload',  HEXDATA, None, None],
                ]
            },
        },
        {
            'id'         : 'vendorMoteReport',
            'name'       : 'vendorMoteReport',
            'description': '',
            'response'   : {
                'FIELDS':  [
                    ['time',     INT,      8,   None],
                    ['macAddr',  STRING,   16,  None],
                    ['payload',  HEXDATA, None, None],
                ]
            },
        },
    ]

    # Notification parsing

    def parse_dataNotif(self, notif_str, notif_fields):
        data_dict = self._parse_xmlobj(notif_str, 'data', notif_fields)
        log.debug('DATA: %s' % str(data_dict))
        # reconvert payload type as a hex value
        data_dict['payloadType'] = int(str(data_dict['payloadType']), 16)
        # set default values for any fields that aren't present
        DEFAULTS = (('isRequest', False),
                    ('isReliable', False),
                    ('isBroadcast', False),
                    ('callbackId', 0))
        for f, v in DEFAULTS:
            if not data_dict[f]:
                data_dict[f] = v
        return (['data'], data_dict)

    def parse_eventNotif(self, notif_str, notif_fields):
        obj_dict = self._parse_xmlobj(notif_str, 'event', None)
        log.debug('EVENT: %s' % str(obj_dict))
        
        event_name = ['event']
        event_dict = {}
        for event_attr, val in obj_dict.items():
            if type(val) is dict:
                event_name += [self.subcommandIdToName(self.NOTIFICATION, event_name, event_attr)]
                subevent_fields = self.getResponseFields(self.NOTIFICATION, event_name)
                event_dict.update(self._xml_parse_fieldset(val, subevent_fields))
            else:
                # we assume that all event fields are defined
                field = [f for f in notif_fields if f.name == event_attr][0]
                event_dict[event_attr] = self._xml_parse_field(val, field)
        return (event_name, event_dict)

    def parse_notif(self, notif_name, notif_str):
        notif_metadata = self.getDefinition(self.NOTIFICATION, notif_name)
        notif_fields = self.getResponseFields(self.NOTIFICATION, notif_name)
        if notif_metadata.has_key('deserializer'):
            deserialize_func = getattr(self, notif_metadata['deserializer'])
            notif_name, notif_dict = deserialize_func(notif_str, notif_fields)
        else:
            notif_dict = self._parse_xmlobj(notif_str, notif_name[0], notif_fields)
        return (notif_name, notif_dict)
    
    # XML-RPC Serializer
    
    def _xmlrpc_format_field(self, field_value, field_metadata):
        if field_metadata[1] == self.HEXDATA:
            return ''.join(['%02X' % b for b in field_value])
        else:
            return field_value

    def default_serializer(self, commandArray, fields):
        cmd_metadata = self.getDefinition(self.COMMAND, commandArray)
        param_list = []
        # for each field in the input parameters, look up the value in cmd_params
        for p in cmd_metadata['request']:
            param_name = p[0]
            # format the parameter by type
            param_list.append(self._xmlrpc_format_field(fields[param_name], p))
        # param_list = [self.format_field(fields[f[0]], f) for f in cmd_metadata['request']]
        return param_list


    # XML-RPC Deserializer

    def _xml_parse_field(self, str_value, field_metadata):
        if field_metadata.format in [self.INT, self.INTS]:
            return int(str_value)
        elif field_metadata.format == self.FLOAT:
            return float(str_value)
        elif field_metadata.format == self.BOOL:
            if str_value.lower() == 'true':
                return True
            else:
                return False
        elif field_metadata.format == self.HEXDATA:
            returnVal = [int(str_value[i:i+2], 16) for i in range(0, len(str_value), 2)]
            return returnVal
        else: 
            return str_value

    def _xml_parse_fieldset(self, obj_dict, fields_metadata):
        'Filter and parse fields in obj_dict'
        filtered_dict = {}
        for field in fields_metadata:
            try:
                field_str = obj_dict[field.name]
                filtered_dict[field.name] = self._xml_parse_field(field_str, field)
            except KeyError:
                # some fields are not always present (especially in Statistics)
                filtered_dict[field.name] = ''
        return filtered_dict

    def _parse_xmlobj(self, xml_doc, base_element, fields_metadata, isArray = False):
        log.debug('Parsing XML: %s %s', base_element, xml_doc)
        aFull_resp = xmlutils.parse_xml_obj(xml_doc, base_element, fields_metadata)
        aRes = []
        for full_resp in aFull_resp : 
            if fields_metadata:
                # parse each field listed in the fields_metadata
                res = self._xml_parse_fieldset(full_resp, fields_metadata)
            else:
                res = full_resp
            if not isArray :
                return res
            aRes.append(res)
        return aRes
    
    def default_deserializer(self, cmd_metadata, xmlrpc_resp):
        resp = {}
        #resp = {'_raw_': xmlrpc_resp}
        resp_fields = self.getResponseFields(self.COMMAND, [cmd_metadata['name']])
        if cmd_metadata['response'].has_key(self.FIELDS):
            # unnamed fields are processed in order
            # note: special case the single return value
            if len(resp_fields) is 1:
                resp[resp_fields[0].name] = self._xml_parse_field(xmlrpc_resp, resp_fields[0])
            else:
                for i, field in enumerate(resp_fields):
                    resp[field.name] = self._xml_parse_field(xmlrpc_resp[i], field)

        elif cmd_metadata['id'] in ['getConfig', 'setConfig'] :
            # default getConfig parser
            # TODO: need an ApiDefinition method to get the response object name
            resp_obj = cmd_metadata['response'].keys()[0]
            isArray = False
            if ('isResponseArray' in cmd_metadata) :
                isArray = cmd_metadata['isResponseArray']
            resp = self._parse_xmlobj(xmlrpc_resp, resp_obj, resp_fields, isArray)
        
        return resp

    def deserialize(self, cmd_name, xmlrpc_resp):
        '''\brief Returns the XML-RPC response as a dict

        \returns A tuple of commandName and the response dictionary, 
                 which contains each of the fields of the response. 
        '''
        cmd_metadata = self.getDefinition(self.COMMAND, cmd_name)
        if cmd_metadata.has_key('deserializer'):
            deserializer = getattr(self, cmd_metadata['deserializer'])
        else:
            deserializer = self.default_deserializer
        resp = deserializer(cmd_metadata, xmlrpc_resp)
        return resp
    
    # ----------------------------------------------------------------------
    # Command-specific serialization methods
    # (must be defined ahead of commands)

    def serialize_getConfig(self, commandArray, cmd_params):
        '''\brief Returns an array of parameters for a typical getConfig query
        '''
        cmd_metadata = self.getDefinition(self.COMMAND, commandArray)
        prefix = []
        if 'serializerParam' in cmd_metadata:
            prefix = cmd_metadata['serializerParam']
        config_query = xmlutils.dict_to_xml(cmd_params, prefix)
        return ['all', config_query]

    def serialize_getNetworkStats(self, commandArray, cmd_params):
        stat_query = self._build_stat_set(cmd_params['period'], cmd_params['index'])
        return ['all', '<config><Network><Statistics>%s</Statistics></Network></config>' % stat_query]

    def serialize_getMote(self, commandArray, cmd_params):
        config_doc = '<config><Motes><Mote><macAddr>%s</macAddr></Mote></Motes></config>' % (cmd_params['macAddr'])
        return ['all', config_doc]

    def serialize_getMoteStats(self, commandArray, cmd_params):
        stat_query = self._build_stat_set(cmd_params['period'], cmd_params['index'])
        config_doc = '<config><Motes><Mote><macAddr>%s</macAddr><Statistics>%s</Statistics></Mote></Motes></config>' % (cmd_params['macAddr'], stat_query)
        return ['all', config_doc]

    def serialize_getPath(self, commandArray, cmd_params):
        config_doc = '<config><Paths><Path><moteMac>%s</moteMac></Path></Paths></config>' % (cmd_params['moteMac'])
        return ['all', config_doc]
    
    def serialize_getPathStats(self, commandArray, cmd_params):
        stat_query = self._build_stat_set(cmd_params['period'], cmd_params['index'])
        config_doc = '<config><Paths><Path><pathId>%s</pathId><Statistics>%s</Statistics></Path></Paths></config>' % (cmd_params['pathId'], stat_query)
        return ['all', config_doc]

    def serialize_getUser(self, commandArray, cmd_params):
        config_doc = '<config><Users><User><userName>%s</userName></User></Users></config>' % (cmd_params['userName'], )
        return ['all', config_doc]


    def _configDoc_format_field(self, field_value, field_metadata):
        if field_metadata[1] == self.HEXDATA:
            return ''.join(['%02X' % b for b in field_value])
        elif field_metadata[1] == self.BOOL:
            return 'true' if field_value else 'false'
        else:
            return str(field_value)

    def serialize_setConfig(self, commandArray, fields) :
        cmd_metadata = self.getDefinition(self.COMMAND, commandArray)
        prefix = []
        if 'serializerParam' in cmd_metadata : 
            prefix = cmd_metadata['serializerParam']
        param_dict = {}
        # for each field in the input parameters, look up the value in cmd_params
        for p in cmd_metadata['request']:
            param_name = p[0]
            # format the parameter by type
            param_dict[param_name] = self._configDoc_format_field(fields[param_name], p)
        config_doc = xmlutils.dict_to_xml(param_dict, prefix)
        return [config_doc]

    def serialize_setBlacklist(self, commandArray, fields):
        cmd_metadata = self.getDefinition(self.COMMAND, commandArray)
        prefix = []
        if 'serializerParam' in cmd_metadata : 
            prefix = cmd_metadata['serializerParam']
        params = fields['frequency'].split()        
        return [xmlutils.list_to_xml(params, 'frequency', prefix)]

    def _build_stat_set(self, period, index = 0):
        STAT_PERIOD_QUERY_TMPL = '<{0}Set><{0}><index>{1}</index></{0}></{0}Set>'
        if period in ['current']:
            return '<statCur/>'
        elif period in ['lifetime']:
            return '<lifetime/>'
        elif period in ['short']:
            return STAT_PERIOD_QUERY_TMPL.format('stat15Min', index)
        elif period in ['long']:
            return STAT_PERIOD_QUERY_TMPL.format('stat1Day', index)
        else:
            raise RuntimeError('invalid stat period: %s', period)
    
    # ----------------------------------------------------------------------
    # Command-specific deserialization methods
    # (must be defined ahead of commands)

    def deserialize_getStats(self, cmd_metadata, xmlrpc_resp):
        net_stats = {}
        fields = self.getResponseFields(self.COMMAND, [cmd_metadata['name']])
        # parse the Statistics element
        resp_dict = self._parse_xmlobj(xmlrpc_resp, 'Statistics', None)
        # detect the statistics period
        stat_dict = {}
        if 'statCur' in resp_dict:
            stat_dict = resp_dict['statCur']
        elif 'lifetime' in resp_dict:
            stat_dict = resp_dict['lifetime']
        elif 'stat15MinSet' in resp_dict:
            stat_dict = resp_dict['stat15MinSet']['stat15Min']
        elif 'stat1DaySet' in resp_dict:
            stat_dict = resp_dict['stat1DaySet']['stat1Day']
        # if there are no statistics for the requested period, stat_dict may be a string 
        if type(stat_dict) != dict:
            stat_dict = {}
        # fill in the statistics fields
        for field in fields:
            try:
                field_str = stat_dict[field.name]
                net_stats[field.name] = self._xml_parse_field(field_str, field)
            except KeyError:
                # some fields are not always present (especially in Statistics)
                net_stats[field.name] = ''  # default value
        return net_stats

    def deserialize_getSourceRoute(self, cmd_metadata, xmlrpc_resp):
        fields = self.getResponseFields(self.COMMAND, [cmd_metadata['name']])
        resp_dict = self._parse_xmlobj(xmlrpc_resp, 'SourceRoute', None)
        # Ug. deserialization for this case is heavily dependant on response structure
        for path in ['primaryPath', 'secondaryPath']:
            if path in resp_dict and 'macAddr' in resp_dict[path]:
                resp_dict[path] = resp_dict[path]['macAddr']
            else:
                resp_dict[path] = []            
        return resp_dict

    def _parse_alarm(self, alarm_dict, alarm_fields, alarm_types):
        '''Reformat a single Alarm object as a flat list of fields'''
        field_names = [f.name for f in alarm_fields]
        # ensure all expected fields are initialized
        result = {f: '' for f in field_names} 
        for f, v in alarm_dict.items():
            # detect the element identifying the alarm type
            if f in alarm_types:
                result['alarmType'] = f
                # this value may contain a dict of alarm-specific data
                if type(v) is dict:
                    # for any matching fields in the alarm-specific data,
                    # flatten them into the result
                    for f2, v2 in v.items():
                        if f2 in field_names:
                            result[f2] = v[f2]
            elif f in field_names:
                result[f] = v
        return result

    def deserialize_getOpenAlarms(self, cmd_metadata, xmlrpc_resp):
        fields = self.getResponseFields(self.COMMAND, [cmd_metadata['name']])
        resp_dict = self._parse_xmlobj(xmlrpc_resp, 'Alarms', None)
        # this parses the list of XML Alarms as:
        # {'Alarm': [{'eventId':...}, {'eventId':...}]}
        # but if there's only one Alarm:
        # {'Alarm': {'eventId':...}}
        # Ug. deserialization for this case is heavily dependant on Alarm types
        result = []
        if 'Alarm' in resp_dict:
            alarm_types = self.getResponseFieldOptions(self.COMMAND, [cmd_metadata['name']], 'alarmType').validOptions
            if type(resp_dict['Alarm']) is list:
                # use the list of Alarms
                alarms = resp_dict['Alarm']
            elif type(resp_dict['Alarm']) is dict:
                # put the single Alarm into a list
                alarms = [resp_dict['Alarm']]
            else:
                raise RuntimeError('invalid Alarm response: can not find Alarm element')
            result = [self._parse_alarm(alarm, fields, alarm_types)
                      for alarm in alarms]            
        return result

    def deserialize_blacklist(self, cmd_metadata, xmlrpc_resp):
        # the same deserializer is used for getConfig and setConfig operations
        resp_fields = self.getResponseFields(self.COMMAND, [cmd_metadata['name']])
        resp_obj = cmd_metadata['response'].keys()[0]
        # cmd_metadata['isResponseArray'] should be True
        resp_dict = xmlutils.parse_xml_obj(xmlrpc_resp, resp_obj, resp_fields)[0]
        # resp_dict['frequency'] contains either a single string or a list of string values
        if type(resp_dict['frequency']) is list:
            resp = [{'frequency': int(freq)} for freq in resp_dict['frequency']]
        else:
            resp = [{'frequency': int(resp_dict['frequency'])}]
        # we return a list of objects containing frequency values
        return resp

    # Commands
    commands = [
        # Get Config commands
        # TODO: use serializer_getConfig instead of command-specific serializers
        
        # activateAdvertising
        {
            'id'         : 'activateAdvertising',
            'name'       : 'activateAdvertising',
            'description': 'Activate advertisement frame',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['timeout',                 INT,      4,   None]
            ],
            'response'   : {
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # activateFastPipe
        {
            'id'         : 'activateFastPipe',
            'name'       : 'activateFastPipe',
            'description': 'Activate the fast network pipe to the specified mote.',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['pipeDirection',           STRING,   25,  'pipeDirection'],
            ],
            'response'   : {
                FIELDS : [
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # cancelOtap
        {
            'id'         : 'cancelOtap',
            'name'       : 'cancelOtap',
            'description': 'This command cancels the OTAP (Over-The-Air-Programming) process to upgrade software on motes and the access point.',
            'request'    : [
            ],
            'response'   : { 
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # cli (deprecated)
        {
            'id'         : 'cli',
            'name'       : 'cli',
            'description': 'This command tunnels a given command through to the manager\'s Command Line Interface (CLI). The CLI command can be called by only one XML API client at a time. The response to the given CLI command is tunneled back to the client via the notifications channel. To receive the CLI notification, the client must be subscribed to CLI notifications (see Notification Channel)',
            'request'    : [
                ['command',                 STRING,   128, None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # deactivateFastPipe
        {
            'id'         : 'deactivateFastPipe',
            'name'       : 'deactivateFastPipe',
            'description': 'Deactivate the fast network pipe to the specified mote.',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # deleteConfig.deleteAcl
        {
            'id'         : 'deleteConfig',
            'name'       : 'deleteAcl',
            'description': 'Remove a device from the ACL',
            'request'    : [
                ['macAddr', STRING, 25, None],
            ],
            'response'   : { 
                FIELDS:  [
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Security', 'Acl', 'Device'],
        },
        # deleteConfig.deleteMote
        {
            'id'         : 'deleteConfig',
            'name'       : 'deleteMote',
            'description': 'Remove a mote from the manager configuration',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : { 
                FIELDS:  [
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Motes', 'Mote'],
        },
        # deleteConfig.deleteUser
        {
            'id'         : 'deleteConfig',
            'name'       : 'deleteUser',
            'description': 'Remove a user',
            'request'    : [
                ['userName',                STRING,   16,  None],
            ],
            'response'   : { 
                FIELDS:  [
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Users', 'User'],
        },
        # exchangeJoinKey
        {
            'id'         : 'exchangeJoinKey',
            'name'       : 'exchangeJoinKey',
            'description': 'Exchange the common join key',
            'request'    : [
                ['newKey',                  HEXDATA,  16,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # exchangeMoteJoinKey
        {
            'id'         : 'exchangeMoteJoinKey',
            'name'       : 'exchangeMoteJoinKey',
            'description': 'Exchange a mote\'s join key',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['newKey',                  HEXDATA,  16,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # exchangeMoteNetworkId
        {
            'id'         : 'exchangeMoteNetworkId',
            'name'       : 'exchangeMoteNetworkId',
            'description': 'Exchange the network ID for a mote',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['newId',                   INT,      4,   None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # exchangeNetworkKey
        {
            'id'         : 'exchangeNetworkKey',
            'name'       : 'exchangeNetworkKey',
            'description': 'Exchange the network key',
            'request'    : [
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # exchangeNetworkId
        {
            'id'         : 'exchangeNetworkId',
            'name'       : 'exchangeNetworkId',
            'description': 'Exchange the network ID',
            'request'    : [
                ['newId',                   INT,      4,   None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # exchangeSessionKey
        {
            'id'         : 'exchangeSessionKey',
            'name'       : 'exchangeSessionKey',
            'description': 'Exchange a mote\'s session key',
            'request'    : [
                ['macAddrA',                STRING,   25,  None],
                ['macAddrB',                STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # decommissionDevice
        {
            'id'         : 'decommissionDevice',
            'name'       : 'decommissionDevice',
            'description': 'Decommission a device in the network',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # getConfig.getAcl
        {
            'id'         : 'getConfig',
            'name'       : 'getAcl',
            'description': '''Check whether a device is part of the ACL''',
            'request'    : [
                ['macAddr',             STRING,   25,  None],
            ],
            'response'   : { 
                'Acl':  [
                    ['macAddr',             STRING,   25,  None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam' : ['config', 'Security', 'Acl', 'Device'],
        },
        # getConfig.getAcls
        {
            'id'         : 'getConfig',
            'name'       : 'getAcls',
            'description': '''Get the list of devices on the ACL''',
            'request'    : [
            ],
            'response'   : { 
                'Acl':  [
                    ['macAddr',             STRING,   25,  None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam' : ['config', 'Security', 'Acl'],
            'isResponseArray': True,
        },
        # getConfig.getBlacklist (returns LIST)
        {
            'id'         : 'getConfig',
            'name'       : 'getBlacklist',
            'description': 'Get the channel blacklist. The output is a list of the blacklisted frequency values.',
            'request'    : [
            ],
            'response'   : {
                'ChannelBlackList':  [
                    ['frequency',           INT,    4,   None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Network', 'ChannelBlackList'],
            'isResponseArray': True,
            'deserializer': 'deserialize_blacklist',
        },
        # getConfig.getMote
        {
            'id'         : 'getConfig',
            'name'       : 'getMote',
            'description': '',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                'Mote':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   25,  None],
                    ['name',                STRING,   16,  None], 
                    ['state',               STRING,   16,  'moteState'],
                    ['numJoins',            INT,      4,   None],
                    ['joinTime',            INT,      8,   None], # TODO: date time
                    ['reason',              STRING,   16,  None], 
                    ['isAccessPoint',       BOOL,     1,   None],
                    ['powerSource',         STRING,   16,  None], # TODO: enum
                    ['dischargeCurrent',    INT,      4,   None],
                    ['dischargeTime',       INT,      4,   None],
                    ['recoveryTime',        INT,      4,   None],
                    ['enableRouting',       BOOL,     1,   None],
                    ['productName',         STRING,   16,  None], 
                    ['hwModel',             INT,      1,   None],
                    ['hwRev',               INT,      1,   None],
                    ['swRev',               STRING,   16,  None], 
                    ['voltage',             FLOAT,    8,   None],
                    ['numNeighbors',        INT,      4,   None],
                    ['needNeighbor',        BOOL,     1,   None],
                    ['goodNeighbors',       INT,      4,   None],
                    ['allocatedPkPeriod',   INT,      4,   None],
                    ['allocatedPipePkPeriod', INT,    4,   None],
                    ['pipeStatus',          STRING,   4,   'pipeStatus'],
                    ['advertisingStatus',   STRING,   4,   'advertisingStatus'],
                    ['locationTag',         STRING,   16,  'locationTag'],                   
                ],
            },
            'serializer' : 'serialize_getMote',
        },
        # setConfig.setMote
        {
            'id'         : 'setConfig',
            'name'       : 'setMote',
            'description': 'Set mote configuration',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['name',                    STRING,   16,  None],
                ['enableRouting',           BOOL,     1,   None],
            ],
            'response'   : { 
                'Mote':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   25,  None],
                    ['name',                STRING,   16,  None], 
                    ['state',               STRING,   16,  'moteState'],
                    ['numJoins',            INT,      4,   None],
                    ['joinTime',            INT,      8,   None], # TODO: date time
                    ['reason',              STRING,   16,  None], 
                    ['isAccessPoint',       BOOL,     1,   None],
                    ['powerSource',         STRING,   16,  None], # TODO: enum
                    ['dischargeCurrent',    INT,      4,   None],
                    ['dischargeTime',       INT,      4,   None],
                    ['recoveryTime',        INT,      4,   None],
                    ['enableRouting',       BOOL,     1,   None],
                    ['productName',         STRING,   16,  None], 
                    ['hwModel',             INT,      1,   None],
                    ['hwRev',               INT,      1,   None],
                    ['swRev',               STRING,   16,  None], 
                    ['voltage',             FLOAT,    8,   None],
                    ['numNeighbors',        INT,      4,   None],
                    ['needNeighbor',        BOOL,     1,   None],
                    ['goodNeighbors',       INT,      4,   None],
                    ['allocatedPkPeriod',   INT,      4,   None],
                    ['allocatedPipePkPeriod', INT,    4,   None],
                    ['pipeStatus',          STRING,   4,   'pipeStatus'],
                    ['advertisingStatus',   STRING,   4,   'advertisingStatus'],
                    ['locationTag',         STRING,   16,  'locationTag'],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Motes', 'Mote'],
        },
        # getConfig.getMotes (return LIST)
        {
            'id'         : 'getConfig',
            'name'       : 'getMotes',
            'description': '''Get the list of Motes''',
            'request'    : [
            ],
            'response'   : {
                'Mote':  [
                    ['moteId',              INT,      4,   None],
                    ['macAddr',             STRING,   25,  None],
                    ['name',                STRING,   16,  None], 
                    ['state',               STRING,   16,  'moteState'],
                    ['numJoins',            INT,      4,   None],
                    ['joinTime',            INT,      8,   None], # TODO: date time
                    ['reason',              STRING,   16,  None], 
                    ['isAccessPoint',       BOOL,     1,   None],
                    ['powerSource',         STRING,   16,  None], # TODO: enum
                    ['dischargeCurrent',    INT,      4,   None],
                    ['dischargeTime',       INT,      4,   None],
                    ['recoveryTime',        INT,      4,   None],
                    ['enableRouting',       BOOL,     1,   None],
                    ['productName',         STRING,   16,  None], 
                    ['hwModel',             INT,      1,   None],
                    ['hwRev',               INT,      1,   None],
                    ['swRev',               STRING,   16,  None], 
                    ['voltage',             FLOAT,    8,   None],
                    ['numNeighbors',        INT,      4,   None],
                    ['needNeighbor',        BOOL,     1,   None],
                    ['goodNeighbors',       INT,      4,   None],
                    ['allocatedPkPeriod',   INT,      4,   None],
                    ['allocatedPipePkPeriod', INT,    4,   None],
                    ['pipeStatus',          STRING,   4,   'pipeStatus'],
                    ['advertisingStatus',   STRING,   4,   'advertisingStatus'],
                    ['locationTag',         STRING,   16,  'locationTag'],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Motes'],
            'isResponseArray': True,
        },
        # getConfig.getMoteStatistics
        {
            'id'         : 'getConfig',
            'name'       : 'getMoteStatistics',
            'description': 'Get the Mote Statistics',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['period',                  STRING,   32,  'statPeriod'],
                ['index',                   INT,      4,   None], # TODO: optional
            ],
            'response'   : { 
                'MoteStatistics':  [
                    ['index',               INT,      4,   None],
                    ['startTime',           INT,      8,   None], # milliseconds
                    ['avgLatency',          INT,      4,   None], # milliseconds
                    ['reliability',         FLOAT,    0,   None], # percentage
                    ['numJoins',            INT,      4,   None],
                    ['voltage',             FLOAT,    4,   None], # volts
                    ['chargeConsumption',   INT,      4,   None], # mC
                    ['temperature',         FLOAT,    4,   None], # deg C
                    # added in Manager 4.1.0.2
                    ['numLostPackets',      INT,      4,   None],
                    # added in Manager 4.1.0.11
                    ['latencyToMote',       INT,      4,   None],
                ],
            },
            'serializer' : 'serialize_getMoteStats',
            'deserializer' : 'deserialize_getStats',
        },
        # getConfig.getNetwork
        {
            'id'         : 'getConfig',
            'name'       : 'getNetwork',
            'description': 'Retrieves network configuration parameters',
            'request'    : [
            ],
            'response'   : { 
                'Network':  [
                    ['netName',             STRING,   16,  None],
                    ['networkId',           INT,      4,   None],
                    ['maxMotes',            INT,      4,   None],
                    ['numMotes',            INT,      4,   None],
                    ['optimizationEnable',  BOOL,     1,   None],
                    ['accessPointPA',       BOOL,     1,   None],
                    ['ccaEnabled',          BOOL,     1,   None],
                    ['requestedBasePkPeriod', INT,    4,   None],
                    ['minServicesPkPeriod', INT,      4,   None],
                    ['minPipePkPeriod',     INT,      4,   None],
                    ['bandwidthProfile',    STRING,   16,  'bandwidthProfile'],
                    ['manualUSFrameSize',   INT,      4,   None],
                    ['manualDSFrameSize',   INT,      4,   None],
                    ['manualAdvFrameSize',  INT,      4,   None],
                    ['netQueueSize',        INT,      4,   None],
                    ['userQueueSize',       INT,      4,   None],
                    ['locationMode',        STRING,   16,  'onOff'],
                    # backbone parameters added in 4.1.0.3
                    ['backboneEnabled',     BOOL,     1,   None],
                    ['backboneSize',        INT,      4,   None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Network'],
        },
        # getConfig.getNetworkStatistics
        {
            'id'         : 'getConfig',
            'name'       : 'getNetworkStatistics',
            'description': 'Get the Network Statistics',
            'request'    : [
                ['period',                  STRING,   32,  'statPeriod'],
                ['index',                   INT,      4,   None], # TODO: optional
            ],
            'response'   : { 
                'NetworkStatistics':  [
                    ['index',               INT,      4,   None],
                    ['startTime',           INT,      8,   None], # milliseconds
                    ['netLatency',          INT,      4,   None], # milliseconds
                    ['netReliability',      FLOAT,    0,   None], # percentage
                    ['netPathStability',    FLOAT,    0,   None], # percentage
                    ['lostUpstreamPackets', INT,      4,   None], # lifetime only ?
                ],
            },
            'serializer' : 'serialize_getNetworkStats',
            'deserializer' : 'deserialize_getStats',
        },
        # getConfig.getOpenAlarms
        {
            'id'         : 'getConfig',
            'name'       : 'getOpenAlarms',
            'description': 'Retrieves a list of the open alarms on the Manager',
            'request'    : [
            ],
            'response'   : {
                'Alarm':  [
                    ['timeStamp',               INT,      4,   None],
                    ['eventId',                 INT,      4,   None],
                    ['alarmType',               STRING,   32,  'alarmType'],
                    ['macAddr',                 STRING,   25,  None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Alarms'],
            'deserializer' : 'deserialize_getOpenAlarms',
            'isResponseArray': True,
        },
        # getConfig.getPaths (returns LIST)
        {
            'id'         : 'getConfig',
            'name'       : 'getPaths',
            'description': '''Get the list of Paths to the mote\'s neighbors''',
            'request'    : [
                # the request parameter moteMac matches the XML query for Paths
                ['moteMac',                STRING,    25,  None],
            ],
            'response'   : {
                'Path':  [
                    ['pathId',              INT,      4,   None],
                    ['moteAMac',            STRING,   25,  None],
                    ['moteBMac',            STRING,   25,  None],
                    ['numLinks',            INT,      4,   None],
                    ['pathDirection',       STRING,   16,  'pathDirection'],
                    ['pathQuality',         FLOAT,    0,   None],
                ],
            },
            'serializer' : 'serialize_getPath',
            'isResponseArray': True,
        },
        # getConfig.getPathStatistics
        {
            'id'         : 'getConfig',
            'name'       : 'getPathStatistics',
            'description': 'Get Statistics for a specific Path',
            'request'    : [
                ['pathId',                  INT,      4,   None],
                ['period',                  STRING,   16,  'statPeriod'],
                ['index',                   INT,      4,   None],
            ],
            'response'   : {
                'PathStatistics':  [
                    ['index',               INT,      4,   None],
                    ['startTime',           INT,      8,   None], # milliseconds
                    ['baPwr',               INT,      1,   None],
                    ['abPwr',               INT,      1,   None],
                    ['stability',           FLOAT,    8,   None],
                ],
            },
            'serializer' : 'serialize_getPathStats',
            'deserializer' : 'deserialize_getStats',
        },
        # getConfig.getRedundancy
        {
            'id'         : 'getConfig',
            'name'       : 'getRedundancy',
            'description': 'Get the redundancy state',
            'request'    : [
            ],
            'response'   : {
                'Redundancy':  [
                    ['localMode',            STRING,  16,  'redundancyMode'],
                    ['peerStatus',           STRING,  16,  'redundancyPeerStatus'],
                    ['peerControllerSwRev',  STRING,  16,  None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Redundancy'],
        },
        # getConfig.getSecurity 
        {
            'id'         : 'getConfig',
            'name'       : 'getSecurity',
            'description': '''Get the Security configuration''',
            'request'    : [
            ],
            'response'   : { 
                'Security':  [
                    ['securityMode',        STRING,   20,  'securityMode'],
                    ['acceptHARTDevicesOnly', BOOL,   1,   None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam' : ['config', 'Security'],
        },
        # getConfig.getSla
        {
            'id'         : 'getConfig',
            'name'       : 'getSla',
            'description': 'Get the Service Level Agreement (SLA) configuration',
            'request'    : [
            ],
            'response'   : {
                'Sla':  [
                    ['minNetReliability',   FLOAT,    8,   None],
                    ['maxNetLatency',       INT,      4,   None],
                    ['minNetPathStability', FLOAT,    8,   None],
                    ['apRdntCoverageThreshold', FLOAT,8,   None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Network', 'Sla'],
        },
        # getConfig.getSystem
        {
            'id'         : 'getConfig',
            'name'       : 'getSystem',
            'description': 'Retrieves system-level information',
            'request'    : [
            ],
            'response'   : {
                'System':  [
                    ['systemName',          STRING,   32,  None],
                    ['location',            STRING,   32,  None],
                    ['swRev',               STRING,   32,  None],
                    ['hwModel',             STRING,   32,  None],
                    ['hwRev',               STRING,   32,  None],
                    ['serialNumber',        STRING,   32,  None],
                    ['time',                INT,      8,   None],
                    ['startTime',           INT,      8,   None],
                    ['cliTimeout',          INT,      4,   None],
                    ['controllerSwRev',     STRING,   32,  None],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'System'],
        },
        # getConfig.getUser 
        {
            'id'         : 'getConfig',
            'name'       : 'getUser',
            'description': 'Get the description of a user ',
            'request'    : [
                ['userName',                STRING,   16,  None],
            ],
            'response'   : {
                'User':  [
                    ['userName',            STRING,   16,  None],
                    ['privilege',           STRING,   16,  'userPrivilege'],
                ],
            },
            'serializer' : 'serialize_getUser',
        },
        # getConfig.getUsers (returns LIST)
        {
            'id'         : 'getConfig',
            'name'       : 'getUsers',
            'description': 'Get the list of users',
            'request'    : [
            ],
            'response'   : {
                'User':  [
                    ['userName',            STRING,   16,  None],
                    ['privilege',           STRING,   16,  'userPrivilege'],
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'Users'],
            'isResponseArray': True,
        },
        # getConfig.getSourceRoute
        {
            'id'         : 'getConfig',
            'name'       : 'getSourceRoute',
            'description': 'Get the Source Route for a specific Mote',
            'request'    : [
                ['destMacAddr',             STRING,   25,  None],
            ],
            'response'   : {
                'SourceRoute':  [
                    ['destMacAddr',         STRING,   25,  None],
                    ['primaryPath',         LIST,     16,  None], 
                    ['secondaryPath',       LIST,     16,  None], 
                ],
            },
            'serializer' : 'serialize_getConfig',
            'serializerParam': ['config', 'SourceRoutes', 'SourceRoute'],
            'deserializer' : 'deserialize_getSourceRoute',
        },
        # getLatency
        {
            'id'         : 'getLatency',
            'name'       : 'getLatency',
            'description': 'Get estimated latency for a mote.',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['downstream',          INT,      4,   None],
                    ['upstream',            INT,      4,   None],
                ],
            },
        },
        # getLicense
        {
            'id'         : 'getLicense',
            'name'       : 'getLicense',
            'description': '''Get the software license key.''',
            'request'    : [
            ],
            'response'   : {
                FIELDS : [ 
                    ['license',             STRING,   40,  None],
                ],
            },
        },
        # getTime
        {
            'id'         : 'getTime',
            'name'       : 'getTime',
            'description': 'Get the current time.',
            'request'    : [
            ],
            'response'   : {
                FIELDS : [ 
                    ['utc_time',            FLOAT,    0,   None], # TODO: return as a date-time format
                    ['asn_time',            INT,      8,   None],
                ],
            },
        },
        # pingMote
        {
            'id'         : 'pingMote',
            'name'       : 'pingMote',
            'description': '''Ping the specified mote. A Net Ping Reply event notification will contain the mote's response.''',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # promoteToOperational
        {
            'id'         : 'promoteToOperational',
            'name'       : 'promoteToOperational',
            'description': 'Promote a quarantined device to operational',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # reset
        {
            'id'         : 'reset',
            'name'       : 'reset',
            'description': 'Reset the system or network',
            'request'    : [
                ['object',                  STRING,   25,  'resetObject'],
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # resetWithId
        {
            'id'         : 'reset',
            'name'       : 'resetWithId',
            'description': 'Reset mote by ID',
            'request'    : [
                ['object',                  STRING,   25,  'resetMote'],
                ['moteId',                  INT,      4,   None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # resetWithMac
        {
            'id'         : 'reset',
            'name'       : 'resetWithMac',
            'description': 'Reset mote by MAC address',
            'request'    : [
                ['object',                  STRING,   25,  'resetMote'],
                ['macAddr',                 STRING,   25,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # sendRequest
        {
            'id'         : 'sendRequest',
            'name'       : 'sendRequest',
            'description': 'Send downstream (request) data',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['domain',                  STRING,   16,  'appDomain'],
                ['priority',                STRING,   16,  'packetPriority'],
                ['reliable',                BOOL,     0,   None],
                ['data',                    HEXDATA,  None,None],
            ],
            'response'   : { 
                FIELDS : [
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # sendResponse
        {
            'id'         : 'sendResponse',
            'name'       : 'sendResponse',
            'description': 'Send downstream data as a response. sendResponse should only be used in special cases.',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['domain',                  STRING,   16,  'appDomain'],
                ['priority',                STRING,   16,  'packetPriority'],
                ['reliable',                BOOL,     0,   None],
                ['callbackId',              INT,      4,   None],
                ['data',                    HEXDATA, None, None],
            ],
            'response'   : { 
                FIELDS : [
                    ['callbackId',          INT,      4,   None],
                ],
            },
        },
        # setConfig.setAcl
        {
            'id'         : 'setConfig',
            'name'       : 'setAcl',
            'description': 'Add or update a device in the ACL',
            'request'    : [
                ['macAddr',                 STRING,   25,  None],
                ['joinKey',                 HEXDATA,  16,  None],
            ],
            'response'   : { 
                'Device':  [
                    ['macAddr',             STRING,   25,  None],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Security', 'Acl', 'Device'],
        },
        # setConfig.setBlacklist
        {
            'id'         : 'setConfig',
            'name'       : 'setBlacklist',
            'description': 'Update the channel blacklist. The input is a list of blacklisted frequency values separated by spaces.',
            'request'    : [
                ['frequency',               STRING,   64,   None],
            ],
            'response'   : { 
                'ChannelBlackList':  [
                    ['frequency',           INT,      4,   None],
                ],
            },
            'serializer' : 'serialize_setBlacklist',
            'serializerParam' : ['config', 'Network', 'ChannelBlackList'],
            'isResponseArray': True,
            'deserializer': 'deserialize_blacklist',
        },
        # setConfig.setNetwork
        {
            'id'         : 'setConfig',
            'name'       : 'setNetwork',
            'description': 'Set network configuration',
            'request'    : [
                ['netName',                 STRING,   16,  None],
                ['networkId',               INT,      4,   None],
                ['maxMotes',                INT,      4,   None],
                ['optimizationEnable',      BOOL,     1,   None],
                ['accessPointPA',           BOOL,     1,   None],
                ['ccaEnabled',              BOOL,     1,   None],
                ['requestedBasePkPeriod',   INT,      4,   None],
                ['minServicesPkPeriod',     INT,      4,   None],
                ['minPipePkPeriod',         INT,      4,   None],
                ['bandwidthProfile',        STRING,   16,  'bandwidthProfile'],
                ['manualUSFrameSize',       INT,      4,   None],
                ['manualDSFrameSize',       INT,      4,   None],
                ['manualAdvFrameSize',      INT,      4,   None],
                ['locationMode',            STRING,   8,   'onOff'],
                # backbone parameters added in 4.1.0.3
                # TODO: UG! adding these parameters prevents backward compatibility
                #['backboneEnabled',     BOOL,     1,   None],
                #['backboneSize',        INT,      4,   None],
            ],
            'response'   : { 
                'Network':  [
                    # TODO: return all elements?
                    ['netName',             STRING,   16,  None],
                    ['networkId',           INT,      4,   None],
                    ['maxMotes',            INT,      4,   None],
                    ['optimizationEnable',  BOOL,     1,   None],
                    ['accessPointPA',       BOOL,     1,   None],
                    ['ccaEnabled',          BOOL,     1,   None],
                    ['requestedBasePkPeriod', INT,    4,   None],
                    ['minServicesPkPeriod', INT,      4,   None],
                    ['minPipePkPeriod',     INT,      4,   None],
                    ['bandwidthProfile',    STRING,   16,  'bandwidthProfile'],
                    ['manualUSFrameSize',   INT,      4,   None],
                    ['manualDSFrameSize',   INT,      4,   None],
                    ['manualAdvFrameSize',  INT,      4,   None],
                    ['locationMode',        STRING,   8,   'onOff'],
                    # backbone parameters added in 4.1.0.3
                    ['backboneEnabled',     BOOL,     1,   None],
                    ['backboneSize',        INT,      4,   None],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Network'],
        },
        # setConfig.setSecurity
        {
            'id'         : 'setConfig',
            'name'       : 'setSecurity',
            'description': 'Set security configuration',
            'request'    : [
                ['securityMode',            STRING,   20,  'securityMode'],
                ['commonJoinKey',           HEXDATA,  16,  None],
                ['acceptHARTDevicesOnly',   BOOL,     1,   None],
            ],
            'response'   : { 
                'Security':  [
                    ['securityMode',        STRING,   20,  'securityMode'],
                    ['acceptHARTDevicesOnly', BOOL,   1,   None],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Security'],
        },
        # setConfig.setSla
        {
            'id'         : 'setConfig',
            'name'       : 'setSla',
            'description': 'Set SLA configuration',
            'request'    : [
                ['minNetReliability',       FLOAT,    8,   None],
                ['maxNetLatency',           INT,      4,   None],
                ['minNetPathStability',     FLOAT,    8,   None],
                ['apRdntCoverageThreshold', FLOAT,    8,   None],
            ],
            'response'   : { 
                'Sla':  [
                    ['minNetReliability',   FLOAT,    8,   None],
                    ['maxNetLatency',       INT,      4,   None],
                    ['minNetPathStability', FLOAT,    8,   None],
                    ['apRdntCoverageThreshold', FLOAT, 8,  None],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Network', 'Sla'],
        },
        # setConfig.setSystem
        {
            'id'         : 'setConfig',
            'name'       : 'setSystem',
            'description': 'Set system-level configuration',
            'request'    : [
                ['systemName',              STRING,   16,  None],
                ['location',                STRING,   16,  None],
                ['cliTimeout',              INT,      4,   None],
            ],
            'response'   : { 
                'System':  [
                    # TODO: return all elements?
                    ['systemName',          STRING,   50,  None],
                    ['location',            STRING,   50,  None],
                    ['cliTimeout',          INT,      4,   None],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'System'],
        },
        # setConfig.setUser
        {
            'id'         : 'setConfig',
            'name'       : 'setUser',
            'description': 'Add or update user configuration',
            'request'    : [
                ['userName',                STRING,   16,  None],
                ['password',                STRING,   16,  None],
                ['privilege',               STRING,   16,  'userPrivilege'],
            ],
            'response'   : { 
                'User':  [
                    ['userName',            STRING,   16,  None],
                    ['privilege',           STRING,   16,  'userPrivilege'],
                ],
            },
            'serializer' : 'serialize_setConfig',
            'serializerParam' : ['config', 'Users', 'User'],
        },
        # setLicense
        {
            'id'         : 'setLicense',
            'name'       : 'setLicense',
            'description': 'Set the software license key.',
            'request'    : [
                ['license',                 STRING,   40,  None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
        },
        # startOtap
        {
            'id'         : 'startOtap',
            'name'       : 'startOtap',
            'description': 'This command initiates the OTAP (Over-The-Air-Programming) process to upgrade software on motes and the Access Point. By default, the process will retry the OTAP file transmission 100 times.',
            'request'    : [
            ],
            'response'   : { 
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # startOtapWithRetries 
        {
            'id'         : 'startOtap',
            'name'       : 'startOtapWithRetries',
            'description': 'This command initiates the OTAP (Over-The-Air-Programming) process to upgrade software for motes and the Access Point, using the specified number of retries.',
            'request'    : [
                ['retries',                 INT,      1,   None]
            ],
            'response'   : { 
                FIELDS : [ 
                          ['result',        STRING,   32,  None],
                ],
            },
        },
        # subscribe
        {
            'id'         : 'subscribe',
            'name'       : 'subscribe',
            'description': '''Subscribe to notifications. This function adds or updates the subscribed notifications to match 'filter'. The filter is a space-separated list of notification types. Valid types include 'data' and 'events'.''',
            'request'    : [
                ['filter',                  STRING,   128, None],
            ],
            'response'   : {
                FIELDS : [ 
                    ['notif_token',         STRING,   32,  None],
                    # note: this is intentionally different from the subscribe API command
                    # (not returning the notification port) because the SmartMesh SDK method
                    # also handles connecting to the notification channel.
                ],
            },
            'command_override': 'subscribe_override',
        },
        # unsubscribe
        {
            'id'         : 'unsubscribe',
            'name'       : 'unsubscribe',
            'description': 'Unsubscribe from notifications. This function clears the existing notification subscription of the client and stops the notification thread. ',
            'request'    : [
                # note: this is intentionally different from the unsubscribe API
            ],
            'response'   : {
                FIELDS : [ 
                    ['result',              STRING,   32,  None],
                ],
            },
            'command_override': 'unsubscribe_override',
        },
    ]
